"""
Custom exceptions used by Twitch Archiver.
"""

import logging
import tempfile
from pathlib import Path

import requests

log = logging.getLogger()


# Docstring for exception are used as exception message`.
# reference:
#   https://stackoverflow.com/a/66491013
class TwitchArchiverError(Exception):
    """Subclass exceptions use docstring as default message."""

    def __init__(self, msg=None, *args, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(msg or (self.__doc__, *args))


class RequestError(TwitchArchiverError):
    def __init__(self, url=None, exception=None):
        """
        :param url: url which returned an error
        :type url: str
        :param exception: exception object
        :type exception: Exception
        """
        message = ""
        if url and exception:
            message = f"API request to {url} failed. Error: {exception}"

        super().__init__(message)


class TwitchAPIError(TwitchArchiverError):
    def __init__(self, response=None):
        """
        :param response: request object
        :type response: requests.Response
        """
        message = ""
        if response:
            message = f"Twitch API returned status code {response.status_code}. URL: {response.url}, Response: {response.text}"

        super().__init__(message)


class TwitchAPIErrorForbidden(TwitchAPIError):
    """Twitch API returned 403 Forbidden."""


class TwitchAPIErrorBadRequest(TwitchAPIError):
    """Twitch API returned 400 Bad Request."""


class TwitchAPIErrorNotFound(TwitchAPIError):
    """Twitch API returned 404 Not Found."""


class StreamFetchError(TwitchArchiverError):
    """Error occurred while fetching stream information."""

    def __init__(self, channel=None):
        message = ""
        if channel:
            message = f"Error occurred while fetching information for stream by {channel.name}."

        super().__init__(message)


class StreamArchiveError(TwitchArchiverError):
    """Error occurred while archiving stream video."""


class StreamDownloadError(StreamArchiveError):
    """Error occurred while downloading stream."""

    def __init__(self, message=None):
        super().__init__(message)


class StreamSegmentDownloadError(StreamDownloadError):
    """Error occurred while downloading stream segment."""

    def __init__(self, segment=None, channel=None):
        message = ""
        if segment and channel:
            message = f"Error occurred while downloading stream segment {segment.id} for {channel.name}."

        super().__init__(message)


class UnsupportedStreamPartDuration(StreamDownloadError):
    def __init__(self):
        message = (
            "Multiple parts with unsupported duration found which cannot be accurately combined. "
            "Falling back to VOD archiver only."
        )

        super().__init__(message)


class StreamOfflineError(StreamArchiveError):
    def __init__(self, channel):
        message = f"Requested stream ({channel.name}) is not currently live."

        super().__init__(message)


class VideoArchiveError(TwitchArchiverError):
    """Error occurred while archiving VOD video."""


class VideoDownloadError(VideoArchiveError):
    """Error occurred while downloading VOD video."""


class VideoPartDownloadError(VideoDownloadError):
    """Error occurred while downloading VOD part."""


class VideoMergeError(VideoArchiveError):
    """Error occurred while merging VOD."""


class VideoVerificationError(VideoMergeError):
    """VOD failed verification check."""


class VideoConvertError(VideoMergeError):
    """Error occurred while converting VOD."""


class CorruptPartError(VideoConvertError):
    """Corrupt part(s) found while converting VOD. Delete VOD and re-download if issue persists."""

    def __init__(self, parts=None):
        """
        :param parts: set of corrupt parts
        :type parts: set[Part]
        """
        message = None
        self.parts = parts
        if parts:
            message = (
                f"Corrupt parts found when converting VOD file. Delete VOD and re-download if issue persists. "
                f"Parts: {[p.id for p in parts]}"
            )

        super().__init__(message)


class UnhandledDownloadError(TwitchArchiverError):
    def __init__(self, vod):
        # handle unsynced stream download (no vod id)
        if vod.v_id == 0:
            message = (
                f"An unhandled exception occurred while downloading stream {vod.s_id} by {vod.channel.name}. "
                f"Check stream downloaded correctly and remove lock file "
                f"({Path(tempfile.gettempdir(), 'twitch-archiver', f'.lock.{vod.s_id}-stream-only')})."
            )

        else:
            message = (
                f"An unhandled exception occurred while downloading VOD {vod.v_id} by {vod.channel.name}. "
                f"Check stream downloaded correctly and remove lock file "
                f"({Path(tempfile.gettempdir(), 'twitch-archiver', f'.lock.{vod.s_id}')})."
            )

        super().__init__(message)


class ChatArchiveError(TwitchArchiverError):
    """Error occurred while archiving chat for VOD."""


class ChatDownloadError(ChatArchiveError):
    """Error occurred while downloading VOD chat logs."""


class ChatExportError(ChatArchiveError):
    """Error occurred while exporting VOD chat logs."""

    def __init__(self, error):
        message = f"Chat export failed. Error: {error}"

        super().__init__(message)


class ChannelOfflineError(TwitchArchiverError):
    def __init__(self, channel):
        message = f"{channel.name} is offline."

        super().__init__(message)


class DatabaseError(TwitchArchiverError):
    def __init__(self, error, vod_id=None):
        if vod_id:
            message = f"VOD {vod_id} database query failed."

        else:
            message = "Sqlite database connection failed."

        message = f"{message} Error: {error}"

        super().__init__(message)


class DatabaseQueryError(TwitchArchiverError):
    def __init__(self, error):
        message = f"Error querying database. Error: {error}"

        super().__init__(message)


class VodAlreadyCompleted(TwitchArchiverError):
    def __init__(self, vod):
        message = f"VOD {vod.v_id} has already been completed in the requested formats according to the VOD database."

        super().__init__(message)


class VodUnlockingError(TwitchArchiverError):
    def __init__(self, vod):
        if vod.v_id:
            message = (
                f"Failed to remove lock file for stream {vod.v_id} by {vod.channel.name}. Check stream "
                f"downloaded correctly and remove '.lock.{vod.v_id}' file from config directory."
            )
        else:
            message = (
                f"Failed to remove lock file for stream {vod.v_id} by {vod.channel.name}. Check stream "
                f"downloaded correctly and remove '.lock.{vod.s_id}-stream' file from config directory."
            )

        super().__init__(message)


class VodLockedError(TwitchArchiverError):
    def __init__(self, vod):
        if vod.v_id:
            message = f"Lock file (.lock.{vod.v_id}) already present for VOD {vod.v_id} by {vod.channel.name}."
        else:
            message = f"Lock file (.lock.{vod.s_id}-stream) already present for stream {vod.v_id} by {vod.channel.name}."

        super().__init__(message)



"""
Module for generating, saving and retrieving configuration values.
"""

import configparser
import logging
import os


class Configuration:

    """
    Generation, saving and loading of twitch-archive configuration.
    """

    # reference:
    #   https://stackoverflow.com/questions/6198372/most-pythonic-way-to-provide-global-configuration-variables-in-config-py/
    __conf = {
        "pushbullet_key": "",
    }

    _log = logging.getLogger()

    def generate_config(self, args):
        """
        Generates the required configuration from provided arguments, overwriting loaded settings.

        :param args: dict of arguments to generate config from
        :type args: dict
        """
        self._log.debug("Generating config from provided arguments.")

        # overwrite configuration with passed args (if set) as they have precedence
        for argument in args:
            # don't copy value if it is empty and config value set
            if args.get(argument) == "" and self.get(argument):
                continue

            self.set(argument, args.get(argument))

    def load_config(self, conf_file):
        """
        Loads the settings stored in the configuration ini file.

        :param conf_file: path to configuration file
        """
        # create conf file if it doesn't exist
        if not os.path.isfile(conf_file):
            self._log.debug("Config file not found - creating one now.")
            self.create_config_file(conf_file)

        self._log.debug("Loading config from file.")

        config = configparser.ConfigParser()
        config.read(conf_file)

        # load individual settings from file
        for setting in config["settings"]:
            self.set(setting, config["settings"][setting])

    def create_config_file(self, conf_file):
        """
        Creates a configuration file for the storing of settings.\n

        :param conf_file: path to configuration file
        """
        self._log.debug("Creating directories for configuration file.")
        os.makedirs(conf_file.parent, exist_ok=True)

        config = configparser.ConfigParser()
        config.add_section("settings")
        self._log.debug("Current config: %s", self.get())

        for setting in self.__conf:
            config.set("settings", setting, self.get(setting))

        self._log.debug("Writing config to %s", conf_file)

        with open(conf_file, "w", encoding="utf8") as _f:
            config.write(_f)

    @classmethod
    def set(cls, name, value):
        """
        Set a specified attribute.

        :param name:  name of attribute to change
        :param value: value to set attribute to
        """
        cls.__conf[name] = value

    @classmethod
    def get(cls, name=None):
        """
        Retrieve a specified attribute.

        :param name: name of attribute to retrieve value of - 'None' returns all attributes
        :return: requested value(s), None if value cannot be found
        """
        if name is None:
            return cls.__conf

        if name not in cls.__conf:
            return None

        return cls.__conf[name]

    @classmethod
    def import_conf(cls, conf_dict: dict):
        """
        Imports config keys and values from a provided dict.

        :param conf_dict: dictionary of values to import
        :type conf_dict: dict
        """
        cls.__conf = conf_dict

    @classmethod
    def get_sanitized(cls, name=None):
        """
        Retrieves a specified attribute, sanitizing secrets.

        :param name: name of attribute to retrieve value of - 'None' returns all attributes
        :return: requested value(s)
        """
        configuration = cls.__conf.copy()
        for key in ["pushbullet_key"]:
            if configuration[key]:
                configuration.update({key: 24 * "*" + configuration[key][24:]})

        if name is None:
            return configuration

        return configuration[name]

    # reference:
    #   https://stackoverflow.com/questions/35247900/python-creating-an-ini-or-config-file-in-the-users-home-directory
    def save(self, conf_file, name=None):
        """
        Saves the running configuration to the configuration ini.
        """
        self._log.debug("Saving config variable(s) to ini file.")

        # import saved config
        config = configparser.ConfigParser()
        config.read(conf_file)

        # overwrite all vars
        if name is None:
            # overwrite with running config
            for setting in Configuration.get():
                config.set("settings", setting, str(Configuration.get(setting)))

            # save to disk
            with open(conf_file, "w", encoding="utf8") as _f:
                config.write(_f)

        # overwrite one var
        else:
            config.set("settings", name, str(Configuration.get(name)))

            # save to disk
            with open(conf_file, "w", encoding="utf8") as _f:
                config.write(_f)



"""
Classes related to Downloading methods and managing the downloading of a given VOD.
"""

import logging
import tempfile
import traceback
from pathlib import Path

from twitcharchiver.configuration import Configuration
from twitcharchiver.database import Database, INSERT_VOD
from twitcharchiver.exceptions import VodLockedError
from twitcharchiver.vod import ArchivedVod, Vod


class Downloader:
    """
    Strategy pattern for downloading methods.
    """

    def __init__(self, parent_dir: Path, quiet: bool):
        """
        Class Constructor.

        :param parent_dir: path to output downloaded VOD(s) to
        :param quiet: whether progress should be printed
        """
        self._parent_dir: Path = parent_dir
        self._quiet: bool = quiet

        self.vod = Vod()

        self._log = logging.getLogger()

    def start(self):
        """
        Start downloader functions.
        """
        return

    def merge(self):
        """
        Merge downloaded files.
        """
        return

    def export_metadata(self):
        """
        Export metadata for the stored VOD.
        """
        return

    def cleanup_temp_files(self):
        """
        Delete all temporary files.
        """
        return


class DownloadHandler:
    """
    Handles file locking and database insertion for VOD archiving.
    """

    def __init__(self, vod: ArchivedVod):
        """
        Class constructor.

        :param vod: VOD to be managed by download handler
        :type vod: ArchivedVod
        :raises VodAlreadyCompleted: if VOD is already completed in the requested formats according to the database
        :raises VodLockedError: if VOD is locked by another instance
        """
        self._log = logging.getLogger()

        _conf: dict = Configuration.get()
        self._lock_file = None
        self._config_dir: Path = _conf["config_dir"]
        self._with_database: bool = bool(_conf["channel"])
        self.vod: ArchivedVod = vod

        # build path to lock file based on if vod being archived or not
        if self.vod.v_id == 0:
            self._lock_fp = Path(
                tempfile.gettempdir(),
                "twitch-archiver",
                str(self.vod.s_id) + ".lock-stream",
            )

        else:
            self._lock_fp = Path(
                tempfile.gettempdir(), "twitch-archiver", str(self.vod.v_id) + ".lock"
            )

    def __enter__(self):
        """
        Enable 'with' statement functions. Creates a lock file (if one doesn't already exist) for the VOD.

        :return: self
        """
        # attempt to create lock file
        if self.create_lock():
            raise VodLockedError(self.vod)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Enable 'with' statement functions. Removes the VOD lock file and inserts / updates the VOD database with VOD
        information.
        """
        # attempt to delete lock file
        if self.remove_lock():
            self._log.debug("Failed to remove lock file.")

        # don't bother adding to database if exception occurs
        if isinstance(exc_val, BaseException):
            self._log.error(
                "Exception occurred inside DownloadHandler. %s",
                traceback.format_tb(exc_tb),
            )

        else:
            # add VOD to database if exit not due to exception
            if self._with_database:
                self.insert_into_database()

    def create_lock(self):
        """
        Creates a lock file for a given VOD.

        :return: True on lock creation failure
        :rtype: bool
        """
        try:
            self._lock_file = open(self._lock_fp, "x")
            return False

        except FileExistsError:
            self._log.debug("Lock file exists for VOD %s.", self.vod)
            return True

    def remove_lock(self):
        """
        Removes a given lock file.

        :return: Exception string on failure
        """
        try:
            self._lock_file.close()
            self._lock_fp.unlink()
            return None

        except Exception as exc:
            self._log.debug("Failed to remove lock file for VOD %s. %s", self.vod, exc)
            return exc

    def insert_into_database(self):
        """
        Inserts (or updates) the VOD in the VOD database.
        """
        # check if VOD already in database
        with Database(Path(self._config_dir, "vods.db")) as _db:
            downloaded_vod = ArchivedVod.import_from_db(
                _db.execute_query(
                    "SELECT vod_id,stream_id,created_at,chat_archived,video_archived FROM vods WHERE stream_id IS ?",
                    {"stream_id": self.vod.s_id},
                )
            )

            # if already present update it
            if downloaded_vod:
                # set flags for updating
                self.vod.chat_archived = (
                    self.vod.chat_archived or downloaded_vod.chat_archived
                )
                self.vod.video_archived = (
                    self.vod.video_archived or downloaded_vod.video_archived
                )
                _db.execute_query(INSERT_VOD, self.vod.ordered_db_dict())

            else:
                _db.execute_query(INSERT_VOD, self.vod.ordered_db_dict())



"""
Class for storing various basic objects used for information handling.
"""
import re
from pathlib import Path

import m3u8


class Category:
    """
    A category is used to describe what a Twitch stream is currently streaming. This class stores information
    relating to 'categories'.
    """

    def __init__(self, game: dict = None):
        """
        Class constructor.

        :param game: dictionary of game information retrieved from Twitch
        """
        if game is None:
            game = {"id": 0, "name": ""}

        self.id: int = game["id"]
        self.name: str = game["name"]

        self.slug: str = ""
        self.thumbnail_url: str = ""
        self.display_name: str = ""
        self.type: str = ""

        for _k in game.keys():
            if _k == "slug":
                self.slug = game[_k]
            if _k == "boxArtURL":
                self.thumbnail_url = game[_k]
            if _k == "displayName":
                self.display_name = game[_k]
            if _k == "type":
                self.type = game[_k]

    def __repr__(self):
        """
        Create str from Category.

        :return: Category as a string
        :rtype: str
        """
        return str({"id": self.id, "name": self.name})

    def __eq__(self, other):
        """
        Compares two Category instances.

        :param other: Category to compare against
        :type other: Category
        :return: True if they match
        :rtype: bool
        """
        if isinstance(other, self.__class__):
            return bool(self.id == other.id)
        raise TypeError

    def to_dict(self):
        """
        Generates and returns a dictionary of relevant category information.

        :return: dict of values which make up a category.
        :rtype: dict
        """
        return {
            "id": self.id,
            "slug": self.slug,
            "thumbnail_url": self.thumbnail_url,
            "name": self.name,
            "display_name": self.display_name,
            "type": self.type,
        }


class Chapters:
    """
    Chapters split up an individual VOD into separate 'moments', describing what game or stream category each
    section of the VOD contains. This class stores moments related to a given Twitch VOD.
    """

    def __init__(self, moments: list[dict] = None):
        """
        Class constructor.

        :param moments: moments list retrieved from Twitch API
        """
        self._moments: list[Chapters.Moment] = []

        if moments:
            self._moments = [self.Moment(m) for m in moments]

    def __bool__(self):
        """
        Check if Chapter initialized.

        :return: True if Chapter initialized
        :rtype: bool
        """
        return bool(self._moments)

    def __repr__(self):
        """
        Convert Chapter to string.

        :return: string of Chapter moments
        :rtype: str
        """
        return str(self._moments)

    def __iter__(self):
        """
        Iterate through Chapter Moments.

        :return: Moments as iterable:
        :rtype: collections.Iterable[Moment]
        """
        for _m in self._moments:
            yield _m

    def insert_moment(self, moment):
        """
        Inserts a moment into the chapter.

        :param moment: A moment retrieved from the Twitch API
        :type moment: Moment
        """
        self._moments.append(moment)

    @staticmethod
    def create_chapter_from_category(category: Category, duration: int):
        """
        Converts a Category into a single chapter which spans an entire VOD. This is useful when no chapters
        are included with a VOD - this happens when only one category spans its entire length.

        :param category: A VOD category retrieved from the Twitch API.
        :param duration: length in seconds of the VOD
        :return: A Chapter containing a moment of the category spanning the entire VOD length
        :rtype: Chapters
        """
        # convert category into moment spanning whole duration of VOD
        _moment = Chapters.Moment()
        # 'GAME_CHANGE' is type assigned to chapters by Twitch
        _moment.type = "GAME_CHANGE"
        _moment.description = category.display_name or category.name
        _moment.category = category
        _moment.segment = Segment(0, duration)

        _chapter = Chapters()
        _chapter.insert_moment(_moment)
        return _chapter

    class Moment:
        """
        A moment is a portion of a Twitch VOD containing a game or category and the section of the VOD it belongs
        to. This class stores that information.
        """

        def __init__(self, moment: dict = None):
            """
            Class constructor.

            :param moment: dict of values retrieved from Twitch
            """

            self.id: int = int()
            self.segment: Segment = Segment()
            self.type: str = ""
            self.description: str = ""
            self.category: Category = Category()

            if moment:
                self.id = moment["id"]
                self.segment = Segment(
                    moment["positionMilliseconds"] / 1000,
                    moment["durationMilliseconds"] / 1000,
                )
                self.type = moment["type"]
                self.description = moment["description"]

                if "game" in moment.keys():
                    self.category = Category(moment["game"])

        def __repr__(self):
            """
            Returns Moment as a string.

            :return: str of Moment attributes.
            :rtype: str
            """
            return str(
                {
                    "description": self.description,
                    "type": self.type,
                    "segment": self.segment,
                }
            )

        def __bool__(self):
            return bool(self.id)


class Segment:
    """
    A segment of a video is a portion of it described with a position from the start, and a duration.
    """

    def __init__(self, position: float = 0.0, duration: float = 0.0):
        """
        Class constructor.

        :param position: start position of segment in seconds
        :param duration: length of segment in seconds
        """
        self.position: float = position
        self.duration: float = duration

    def __repr__(self):
        """
        Returns Segment as a string.

        :return: Segment attributes as string.
        :rtype: str
        """
        return str({"position": self.position, "duration": self.duration})


class MpegSegment(Segment):
    """
    MpegSegments are the individual pieces which comprise a Twitch VOD. This class defines the storage and
    provides useful methods for handling them.
    """

    def __init__(
        self,
        segment_id: int() = 0,
        duration: int = 0,
        url: str = "",
        muted: bool = False,
    ):
        """
        Class constructor.

        :param segment_id:
        :param duration:
        :param url:
        :param muted:
        """
        self.muted: bool = muted
        self.id: int = segment_id
        self.url: str = url
        super().__init__(self.id * 10, duration)

    def __repr__(self):
        return str({"id": self.id, "duration": self.duration, "muted": self.muted})

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id
        raise TypeError

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self.id != other.id
        raise TypeError

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.id < other.id
        raise TypeError

    @staticmethod
    def convert_m3u8_segment(segment: m3u8.Segment, base_url: str):
        """
        Derives an MpegSegment from a provided m3u8.Segment instance

        :param segment: m3u8.Segment
        :type segment:
        :param base_url: url directory where segments (00000.ts, 000001.ts, ...) are located online
        :type base_url: str
        :return: segment generated from the provided m3u8 segment
        :rtype: MpegSegment
        """
        return MpegSegment(
            int(re.sub(r".ts|-[a-zA-Z]*.ts", "", segment.uri)),
            segment.duration,
            f"{base_url}{segment.uri}",
            "muted" in segment.uri,
        )

    def id_padded(self):
        """
        Generates a 0-padded ID for the current segment.
        """
        return f"{self.id:05d}"

    def generate_url(self, base_url: str):
        """
        Generates a URL based on the segment's id and whether it is muted or not
        """
        return "".join([base_url, str(self.id), "-muted" if self.muted else "", ".ts"])

    def generate_path(self, base_path: Path):
        """
        Generates a path based on the segment's ID and a provided base path
        """
        return Path(base_path, self.id_padded() + ".ts")



"""
Various utility functions for modifying, retrieving and saving information.
"""

import hashlib
import json
import logging
import multiprocessing
import os
import re
import shutil
import traceback
from datetime import datetime, timezone
from itertools import groupby
from math import ceil, floor
from pathlib import Path
from textwrap import dedent

import requests

from twitcharchiver.twitch import Chapters

log = logging.getLogger()


def build_output_dir_name(title: str, created_at: float, vod_id: int = 0):
    """
    Builds a directory name based on a title, a timestamp and VOD ID.

    :param title: name of VOD or stream
    :type title: str
    :param created_at: timestamp
    :type created_at: float
    :param vod_id: ID of VOD
    :type vod_id: int
    :return: name of output folder for given VOD or stream parameters
    :rtype: str
    """
    # trim titles to 128 characters (leaves ~90 characters for the base directory)
    if len(title) > 128:
        title = title[:128] + "..."

    # trim long stream names based on byte length (ext4 has a max filename size of 255 bytes)
    if len(title.encode("utf-8")) > 192:
        # encode to utf-8, truncate then decode
        title = title.encode("utf-8")[:192].decode("utf-8", "ignore") + "..."

    if vod_id != 0:
        _dir_name = " - ".join(
            [format_timestamp(created_at), sanitize_text(title), str(vod_id)]
        )
    else:
        _dir_name = " - ".join(
            [format_timestamp(created_at), sanitize_text(title), "STREAM_ONLY"]
        )
    return _dir_name


def format_timestamp(timestamp: float):
    """
    Formats a given UTC timestamp to the YYYY-MM-DD_HH-MM-SS format.

    :param timestamp: UTC timestamp to convert
    :return: timestamp in YYYY-MM-DD_HH-MM-SS format
    """
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d_%H-%M-%S")


def export_json(vod_json: dict):
    """Exports all VOD information to a file.

    :param vod_json: dict of vod parameters retrieved from twitch
    """
    with open(
        Path(vod_json["store_directory"], "vod.json"), "w", encoding="utf8"
    ) as json_out_file:
        json_out_file.write(json.dumps(vod_json))


def import_json(vod_json: dict):
    """Imports all VOD information from a file.

    :param vod_json: dict of vod parameters retrieved from twitch
    """
    if Path(vod_json["store_directory"], "vod.json").exists():
        with open(
            Path(vod_json["store_directory"], "vod.json"), "r", encoding="utf8"
        ) as json_in_file:
            return json.loads(json_in_file.read())

    return []


# https://stackoverflow.com/a/43091576
def to_ranges(iterable):
    """Converts a list of integers to iterable sets of (low, high) (e.g [0, 1, 2, 5, 7, 8] -> (0, 2), (5, 5), (7, 8))

    :param iterable: list of integers
    :return: iterable generator of separate integer ranges
    """
    iterable = sorted(set(iterable))
    for key, group in groupby(enumerate(iterable), lambda t: t[1] - t[0]):
        group = list(group)
        yield group[0][1], group[-1][1]


def sanitize_text(string: str):
    """Sanitize a given string removing unwanted characters which aren't allowed in directories, file names.

    :param string: string of characters to sanitize
    :return: sanitized string
    """
    if not string:
        return ""

    return re.sub(r'[/\\:|<>"?*\0-\x1f]', "_", string)


def sanitize_date(date):
    """Removes unwanted characters from a timedate structure.

    :param date: date retrieved from twitch to sanitize
    :return: sanitized date
    """
    for r in (("T", "_"), (":", "-"), ("Z", "")):
        date = date.replace(*r)

    return date


def convert_to_seconds(duration):
    """Converts a given time in the format HHhMMmSSs to seconds.

    :param duration: time in HHhMMmSSs format to be converted
    :return: time in seconds
    """
    duration = duration.replace("h", ":").replace("m", ":").replace("s", "").split(":")

    if len(duration) == 1:
        return int(duration[0])

    if len(duration) == 2:
        return (int(duration[0]) * 60) + int(duration[1])

    if len(duration) == 3:
        return (int(duration[0]) * 3600) + (int(duration[1]) * 60) + int(duration[2])

    return int()


def convert_to_hms(seconds):
    """Converts a given time in seconds to the format HHhMMmSSs.

    :param seconds: time in seconds
    :return: time in HHhMMmSSs format
    """
    minutes = seconds // 60
    hours = minutes // 60

    return f"{hours:02d}h{minutes % 60:02d}m{seconds % 60:02d}s"


def time_since_date(timestamp: float):
    """Returns the time in seconds between a given datetime and now.

    :param timestamp: utc timestamp to compare current datetime to
    :return: the time in seconds since the given date
    """
    created_at = int(timestamp)
    current_time = int(datetime.now(timezone.utc).timestamp())

    return current_time - created_at


def get_time_difference(start_time, end_time):
    """Returns the time in seconds between a given datetime and now.

    :param start_time: start utc timestamp
    :param end_time: end utc timestamp
    :return: the time in seconds:milliseconds between the two datetimes
    """

    return end_time - start_time


def parse_twitch_timestamp(timestamp):
    """
    Parses the provided timestamp depending on its format.

    :param timestamp: timestamp to interpret
    :type timestamp: str
    :return: interpreted timestamp
    :rtype: float
    """
    # older twitch timestamps may include microseconds
    if "." in timestamp:
        return (
            datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%fZ")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )
    else:
        return (
            datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )


def get_latest_version():
    """Fetches the latest release information from GitHub.

    :return: latest version number
    :return: latest release notes
    """
    try:
        _r = requests.get(
            "https://api.github.com/repos/Brisppy/twitch-vod-archiver/releases/latest",
            timeout=10,
        )
        # catch error codes such as 403 in case of rate limiting
        if _r.status_code != 200:
            return "0.0.0", ""
        latest_version = _r.json()["tag_name"].replace("v", "")
        release_notes = _r.json()["body"]

    # return a dummy value if request fails
    except Exception:
        return "0.0.0", ""

    return latest_version, release_notes


# reference:
#   https://stackoverflow.com/a/11887825
def version_tuple(v):
    """
    Convert dotted version number to tuple for comparison.

    :param v: dotted version number
    :return: tuple of version numbers
    """
    return tuple(map(int, (v.split("."))))


def check_update_available(local_version, remote_version):
    """
    Compares two software versions.

    :param local_version: local version in use
    :param remote_version: remote version to compare against
    :return: True if remote version has a higher version number, otherwise False
    """
    # check if local version is 'special', as in a development build or release candidate
    local_version_parts = local_version.split(".")
    if len(local_version_parts) > 3:
        log.warning(
            "Currently using a development or release candidate build. These may be unfinished or contain serious "
            "bugs. Report any issues you encounter to https://github.com/Brisppy/twitch-archiver/issues."
        )
        # update is available if we are using a dev or release candidate build equal to or prior to
        # the latest stable release
        if version_tuple(".".join(local_version_parts[:-1])) <= version_tuple(
            remote_version
        ):
            return True

    elif version_tuple(local_version) < version_tuple(remote_version):
        return True

    return False


def send_push(pushbullet_key, title, body=""):
    """
    Sends a push to an account based on a given pushbullet key.

    :param pushbullet_key: key for destination pushbullet account. 'False' to not send.
    :param title: title to send with push
    :param body: body to send with push
    """
    h = {
        "content-type": "application/json",
        "Authorization": f"Bearer {pushbullet_key}",
    }
    d = {"type": "note", "title": f"[twitch-archiver] {title}", "body": body}

    try:
        _r = requests.post(
            url="https://api.pushbullet.com/v2/pushes",
            headers=h,
            data=json.dumps(d),
            timeout=10,
        )

        if _r.status_code != 200:
            if _r.json()["error"]["code"] == "pushbullet_pro_required":
                log.error(
                    "Error sending push. Likely rate limited (500/month). "
                    "Error %s: %s",
                    _r.status_code,
                    _r.text,
                )

            else:
                log.error("Error sending push. Error %s: %s", _r.status_code, _r.text)

    except Exception as exc:
        log.error("Error sending push. Error: %s", exc)


# reference:
#   https://www.geeksforgeeks.org/compare-two-files-using-hashing-in-python/
def get_hash(file):
    """Retrieves the hash for a given file.

    :param file: path to file to hash
    :return: hash of provided file
    """
    f_hash = hashlib.md5()

    with open(Path(file), "rb") as f:
        while True:
            # read in next chunk
            _d = f.read(65536)

            if not _d:
                break

            f_hash.update(_d)

        return f_hash.hexdigest()


# reference:
#   https://alexwlchan.net/2019/03/atomic-cross-filesystem-moves-in-python/
def safe_move(src_file, dst_file):
    """Atomically moves src_file to dst_file

    :param src_file: source file to copy
    :param dst_file: path to copy file to
    :raises FileNotFoundError: if src_file does not exist
    """
    log.debug('Moving "%s" to "%s".', src_file, dst_file)

    if Path(src_file).exists():
        # remove source file if it matches destination file
        dst_exists = os.path.exists(dst_file)
        if dst_exists and os.path.samefile(src_file, dst_file):
            log.debug("%s already exists and matches %s.", dst_file, src_file)
            os.remove(src_file)
        else:
            if dst_exists:
                os.remove(dst_file)

            # generate temp file path and copy source file to it
            tmp_file = Path(Path(dst_file.parent), os.urandom(6).hex())
            shutil.copyfile(src_file, tmp_file)

            # rename temp file
            os.rename(tmp_file, dst_file)

            # delete source file
            os.remove(src_file)

    else:
        raise FileNotFoundError


def getenv(name, default_val=None, is_bool=False):
    """
    Wrapper around os.getenv to convert empty strings to None type

    :param name: environment variable name
    :param default_val: default value to return if environment variable does not exist
    :param is_bool: handle environment variable as a case-insensitive boolean string ('true' or 'false')
    :return: environment variable value
    """
    val = os.getenv(name, default_val)

    if is_bool and isinstance(val, str):
        if val.upper() == "TRUE":
            return True

        if val.upper() == "FALSE":
            return False

        raise ValueError(
            f"Invalid boolean value (true or false) received for environment variable: {name}={val}"
        )

    # default return
    return val


def format_vod_chapters(chapters: Chapters):
    """Formats vod chapters retrieved from Twitch into an FFmpeg insertable format

    :param chapters: Chapters which make up the VOD
    :type chapters: Chapters
    :return: chapters formatted as a string readable by ffmpeg
    """
    formatted_chapters = ";FFMETADATA1\n"
    chapter_base = dedent(
        """\
    [CHAPTER]
    TIMEBASE=1/1000
    START={start}
    END={end}
    title={title}
    
    """
    )

    # some chapters have no game attached and so the 'description' is used instead
    for chapter in chapters:
        formatted_chapters += chapter_base.format(
            start=chapter.segment.position * 1000,
            end=(chapter.segment.position + chapter.segment.duration) * 1000,
            title=chapter.description,
        )

    return formatted_chapters


def write_file(data: str, file: Path):
    """
    Writes data to the provided file.

    :param data: string to write to file
    :type data: str
    :param file: Path of output file (will be overwritten)
    :type file: Path
    """
    try:
        with open(Path(file), "w", encoding="utf8") as _f:
            _f.write(data)

    except Exception as exc:
        log.error('Failed to write data to "%s". Error: %s', Path(file), exc)


def write_file_line_by_line(data: list, file: Path):
    """
    Writes data to the provided file with each list element on a new line.

    :param data: list to write to file
    :type data: list
    :param file: Path of output file (will be overwritten)
    :type file: Path
    """
    try:
        # delete existing file as we must append to it
        if Path(file).is_file():
            Path(file).unlink()

        # write each message line by line to readable log
        with open(Path(file), "a+", encoding="utf-8") as _f:
            for _element in data:
                _f.write(f"{_element}\n")

    except Exception as exc:
        log.error('Failed to write data to "%s". Error: %s', Path(file), exc)


def write_json_file(data, file: Path):
    """
    Writes data to the provided file.

    :param data: dict to write to file
    :type data: dict | list
    :param file: Path of output file (will be overwritten)
    :type file: Path
    """
    try:
        with open(Path(file), "w", encoding="utf8") as _f:
            _f.write(json.dumps(data, default=str))

    except Exception as exc:
        log.error('Failed to write json data to "%s". Error: %s', Path(file), exc)


class Progress:
    """
    Functions for displaying progress.
    """

    start_time = 0

    def __init__(self):
        """
        Sets the start time used for computing the time remaining.
        """
        if self.start_time == 0:
            self.start_time = int(datetime.now(timezone.utc).timestamp())

    # reference:
    #   https://stackoverflow.com/questions/63865536/how-to-convert-seconds-to-hhmmss-format-without-failing-if-hour-more-than-24
    @staticmethod
    def to_hms(s: int):
        """Converts a given time in seconds to HHhMMmSSs.

        :param s: time in seconds
        :return: time formatted as HHhMMmSSs
        """
        m, s = divmod(s, 60)
        h, m = divmod(m, 60)
        return f"{h:0>2}:{m:0>2}:{s:0>2}"

    def print_progress(self, cur: int, total: int):
        """Prints and updates a nice progress bar.

        :param cur: current progress out of total
        :param total: highest value of progress bar
        """
        last_frame = cur == total
        percent = floor(100 * (cur / total))
        progress = floor((0.25 * percent)) * "#" + ceil(25 - (0.25 * percent)) * " "
        if cur == 0 or self.start_time == 0:
            remaining_time = "?"

        else:
            remaining_time = self.to_hms(
                ceil(
                    (
                        (int(datetime.now(timezone.utc).timestamp()) - self.start_time)
                        / cur
                    )
                    * (total - cur)
                )
            )

        if len(str(percent)) < 3:
            percent = " " * (3 - len(str(percent))) + str(percent)

        if len(str(cur)) < len(str(total)):
            cur = " " * (len(str(total)) - len(str(cur))) + str(cur)

        # end with newline rather than return
        if last_frame:
            print(
                f"  100%  -  [#########################]  -  {cur} / {total}  -  ETA: 00:00:00",
                end="\n",
            )

        else:
            print(
                f"  {percent}%  -  [{progress}]  -  {cur} / {total}  -  ETA: {remaining_time}",
                end="\r",
            )



"""
Module used for downloading the video for a given Twitch VOD.
"""
import json
import logging
import os
import re
import shutil
import subprocess
import tempfile
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from math import floor
from pathlib import Path
from time import sleep

import m3u8
import requests
from requests import adapters

from twitcharchiver.api import Api
from twitcharchiver.downloader import Downloader
from twitcharchiver.exceptions import (
    VideoPartDownloadError,
    TwitchAPIErrorNotFound,
    TwitchAPIErrorForbidden,
    VideoDownloadError,
    VideoConvertError,
    CorruptPartError,
    VideoMergeError,
    VideoVerificationError,
)
from twitcharchiver.twitch import MpegSegment
from twitcharchiver.utils import (
    Progress,
    safe_move,
    build_output_dir_name,
    get_hash,
    format_vod_chapters,
    time_since_date,
    write_json_file,
)
from twitcharchiver.vod import Vod, ArchivedVod

# time in seconds between checking for new VOD parts if VOD is currently live and being updated
CHECK_INTERVAL = 60


class Video(Downloader):
    """
    Class used for downloading the video for a given Twitch VOD.
    """

    # class vars
    _api: Api = Api()
    _quality: str = ""

    def __init__(
        self,
        vod: Vod,
        parent_dir: Path = os.getcwd(),
        quality: str = "best",
        threads: int = 20,
        quiet: bool = False,
    ):
        """
        Class used for downloading the video for a given Twitch VOD.

        :param vod: VOD to be downloaded
        :param parent_dir: path to parent directory for downloaded files
        :param quality: quality to download in the format [resolution]p[framerate], or either 'best' or 'worst'
        :param quiet: boolean whether to print progress
        :param threads: number of worker threads to use when downloading
        """
        # init downloader
        super().__init__(parent_dir, quiet)

        self.threads = threads

        # set quality
        self.__setattr__("_quality", quality)

        # vod-specific vars
        self.vod: Vod = vod
        self.output_dir = Path(
            self._parent_dir,
            build_output_dir_name(self.vod.title, self.vod.created_at, self.vod.v_id),
        )

        # buffers and progress tracking
        # collect previously downloaded segments (if any)
        self._completed_segments: set[MpegSegment] = self.get_completed_segments(
            self.output_dir
        )
        self._muted_segments: set[MpegSegment] = set()

        # expand download https session pool
        self._s: requests.Session = requests.session()
        _a = requests.adapters.HTTPAdapter(pool_connections=100, pool_maxsize=100)
        self._s.mount("https://", _a)

        # video segment containers and required params
        self._index_url: str = ""
        self._base_url: str = ""
        self._index_playlist: m3u8 = None
        self._prev_index_playlist: m3u8 = None

    def export_metadata(self):
        write_json_file(self.vod.to_dict(), Path(self.output_dir, "vod.json"))

    @staticmethod
    def get_completed_segments(directory):
        """
        Fetches all segments which have been downloaded to the output directory.

        :param directory: directory containing *.ts files
        :type directory: str or Path
        :return: set of segments inside directory
        :rtype: set[MpegSegment]
        """
        return {
            MpegSegment(int(Path(p).name.removesuffix(".ts")), 10)
            for p in list(Path(directory, "parts").glob("*.ts"))
        }

    def start(self, _q=None):
        """
        Begin downloading video segments for given VOD until all parts downloaded and stream has ended (if live).

        :param _q: multiprocessing queue used for returning the class after completion
        :type _q: multiprocessing.Queue
        """
        try:
            # create output directories
            Path(self.output_dir, "parts").mkdir(parents=True, exist_ok=True)
            Path(tempfile.gettempdir(), "twitch-archiver", str(self.vod.v_id)).mkdir(
                parents=True, exist_ok=True
            )

            # delay archival start for new vods (started less than 5m ago)
            _time_since_start = self.vod.time_since_live()
            if _time_since_start < 300:
                self._log.info(
                    "Delaying archival of VOD %s as VOD has been live for less than 5 minutes.",
                    self.vod.v_id,
                )
                sleep(300 - _time_since_start)

            self._index_url = self.vod.get_index_url(self._quality)
            self._base_url = self._extract_base_url(self._index_url)

            # download once
            self._download()

            # while VOD live, check for new parts every CHECK_INTERVAL seconds. if no new parts discovered after CHECK_INTERVAL * VOD_OFFLINE_TIME seconds, or error
            # returned when trying to check VOD status, stream is assumed to be offline and we break loop.
            while self.vod.is_live():
                self._log.debug(
                    "VOD is still live, attempting to download new video segments."
                )
                _start_timestamp: float = datetime.now(timezone.utc).timestamp()

                # refresh VOD metadata
                self.vod.refresh_vod_metadata()
                self._download()

                # sleep if processing time < CHECK_INTERVAL before fetching new messages
                _loop_time = int(
                    datetime.now(timezone.utc).timestamp() - _start_timestamp
                )
                if _loop_time < CHECK_INTERVAL:
                    sleep(CHECK_INTERVAL - _loop_time)

            # delay final archive pass if stream just ended
            self.vod.refresh_vod_metadata()
            if time_since_date(self.vod.created_at + self.vod.duration) < 300:
                self._log.debug(
                    "Stream ended less than 5m ago, delaying before final video archive attempt."
                )
                sleep(300)

                # refresh VOD metadata
                self.vod.refresh_vod_metadata()
                self._download()

            # set archival flag if ArchivedVod provided
            if isinstance(self.vod, ArchivedVod):
                self.vod.video_archived = True

        except (TwitchAPIErrorNotFound, TwitchAPIErrorForbidden):
            self._log.warning(
                "HTTP code 403 or 404 encountered, VOD %s was likely deleted.",
                self.vod.v_id,
            )
            with open(
                Path(self.output_dir, ".ignorelength"), "w", encoding="utf8"
            ) as _:
                pass

            self.vod.status = "offline"

        except Exception as exc:
            raise VideoDownloadError(exc) from exc

        finally:
            # put self into mp queue if provided
            if _q:
                _q.put(self, block=False)

    def refresh_playlist(self):
        """
        Fetch new segments for video (if any).
        """
        self._prev_index_playlist = self._index_playlist
        _raw_playlist = self.vod.get_index_playlist(self._index_url)
        self._index_playlist = m3u8.loads(_raw_playlist)
        # update VOD duration
        try:
            self.vod.duration = floor(
                float(
                    re.findall(
                        r"(?<=#EXT-X-TWITCH-TOTAL-SECS:).*(?=\n)", _raw_playlist
                    )[0]
                )
            )
        except Exception as exc:
            self._log.error("Failed to update VOD duration. Error: %s", exc)

    def _download(self):
        """
        Begin downloading segments, fetching new segments as they come out (if VOD live) until stream/VOD ends.
        """
        self.refresh_playlist()
        self.download_m3u8_playlist()

        # refresh mpegts segment playlist
        self.refresh_playlist()

        # fetch downloaded files in-case being run in parallel with stream archiver so we don't try and
        # download anything already completed
        self._completed_segments = self.get_completed_segments(self.output_dir)

        # if new segments found, download them
        if len(self._prev_index_playlist.segments) < len(self._index_playlist.segments):
            self._log.debug("New VOD parts found.")
            self.download_m3u8_playlist()

    @staticmethod
    def _extract_base_url(index_url: str):
        """
        Extracts a URL from which TS segment IDs are appended to generate the URL which segments are stored.

        :param index_url: index url used to create base url
        :return: base url for TS segments
        """
        _m = re.findall(r"(?<=\/)(index.*)", index_url)[0]
        return index_url.replace(_m, "")

    def download_m3u8_playlist(self):
        """Downloads the video for a specified m3u8 playlist.

        :raises vodPartDownloadError: error returned when downloading vod parts
        """
        _buffer: set[MpegSegment] = set()

        # process all segments in playlist
        for segment in [
            MpegSegment.convert_m3u8_segment(_s, self._base_url)
            for _s in self._index_playlist.segments
        ]:
            # add segment to download buffer if it isn't completed
            if segment not in self._completed_segments:
                _buffer.add(segment)

            if segment.muted:
                self._muted_segments.add(segment)

        if _buffer:
            _worker_pool = ThreadPoolExecutor(max_workers=self.threads)
            download_error = []
            futures = []
            try:
                # add orders to worker pool
                for segment in _buffer:
                    futures.append(_worker_pool.submit(self._get_ts_segment, segment))

                progress = Progress()

                # complete orders in worker pool
                for future in futures:
                    if future.exception():
                        # append any returned errors
                        download_error.append(future.exception())

                    if not self._quiet:
                        progress.print_progress(
                            len(self._completed_segments),
                            len(self._index_playlist.segments),
                        )

                if download_error:
                    raise VideoPartDownloadError(download_error)

            except KeyboardInterrupt as exc:
                self._log.debug(
                    "M3U8 playlist downloader caught interrupt, shutting down workers..."
                )
                _worker_pool.shutdown(wait=False, cancel_futures=True)
                raise KeyboardInterrupt from exc

            finally:
                _worker_pool.shutdown(wait=False, cancel_futures=True)

    def _get_ts_segment(self, segment: MpegSegment):
        """Retrieves a specific ts file.

        :param segment: MPEGTS segment to download
        :type segment: Segment
        :return: error on failure
        :rtype: str
        """
        _segment_path = segment.generate_path(Path(self.output_dir, "parts"))
        self._log.debug("Downloading segment %s to %s", segment.url, _segment_path)

        # don't bother if piece already downloaded
        if os.path.exists(_segment_path):
            return

        # files are downloaded to $TMP, then moved to final destination
        # takes 3:32 to download an hour long VOD to NAS, compared to 5:00 without using $TMP as download cache
        #   a better method would be to have 20 workers downloading, and 20 moving temp
        #   files from storage avoiding any downtime downloading

        # create temporary file for downloading to
        with open(
            Path(
                tempfile.gettempdir(),
                "twitch-archiver",
                str(self.vod.v_id),
                f"{segment.id}.ts",
            ),
            "wb",
        ) as _tmp_ts_file:
            for _ in range(6):
                if _ > 4:
                    raise VideoPartDownloadError(
                        f"Maximum retries for segment {segment.id} reached."
                    )

                try:
                    _r = self._s.get(segment.url, stream=True, timeout=10)

                    # break on non 200 status code
                    if _r.status_code != 200:
                        self._log.error(
                            "HTTP status %s received. %s", _r.status_code, _r.text
                        )
                        continue

                    # write downloaded chunks to temporary file
                    for _chunk in _r.iter_content(chunk_size=262144):
                        _tmp_ts_file.write(_chunk)

                    self._log.debug(
                        "Segment %s download completed.", Path(_segment_path).stem
                    )
                    self._completed_segments.add(segment)

                    break

                except requests.exceptions.RequestException as exc:
                    self._log.debug(
                        "Segment %s download failed (Attempt %s). Error: %s)",
                        Path(_segment_path).stem,
                        _ + 1,
                        str(exc),
                    )
                    continue

        try:
            # move part to destination storage
            safe_move(_tmp_ts_file.name, _segment_path)
            self._log.debug(
                "Segment %s completed and moved to %s.",
                Path(_segment_path).stem,
                _segment_path,
            )

        except FileNotFoundError as exc:
            raise VideoPartDownloadError(
                f"MPEG-TS segment did not download correctly. Piece: {segment.url}"
            ) from exc

        except Exception as exc:
            raise VideoPartDownloadError(
                f"Exception encountered while moving downloaded MPEG-TS segment {segment.id}."
            ) from exc

    def repair_vod_corruptions(self, corruption: set[MpegSegment]):
        """
        Repairs segments which have been declared as corrupt by VOD merger.

        :param corruption: set of corrupt MpegSegments retrieved from CorruptVodError exception.
        """
        # check vod still available
        if not self._index_url:
            raise VideoDownloadError(
                "Corrupt segments were found while converting VOD and TA was unable to re-download the missing "
                "segments. Either re-download the VOD if it is still available, or manually convert 'merged.ts' using "
                f"FFmpeg. Corrupt parts:\n{str(sorted(corruption))}"
            )

        # rename corrupt segments
        for segment in corruption:
            # convert segment number to segment file
            segment_fp = str(f"{segment.id:05d}" + ".ts")

            # rename part
            shutil.move(
                Path(self.output_dir, "parts", segment_fp),
                Path(self.output_dir, "parts", segment_fp + ".corrupt"),
            )

            # remove from completed segments
            self._completed_segments.remove(segment)

        # download and combine vod again
        try:
            self.refresh_playlist()
            self.download_m3u8_playlist()

            # compare downloaded .ts to corrupt parts - corrupt parts SHOULD have different hashes,
            # so we can work out if a segment is corrupt on twitch's end or ours
            for segment in corruption:
                segment_fp = str(f"{segment.id:05d}" + ".ts")

                # compare hash of redownloaded segment and corrupt one
                try:
                    if get_hash(Path(self.output_dir, "parts", segment_fp)) == get_hash(
                        Path(self.output_dir, "parts", segment_fp + ".corrupt")
                    ):
                        self._log.debug(
                            "Re-downloaded .ts segment %s matches corrupt one, "
                            "assuming corruption is on Twitch's end and ignoring.",
                            segment.id,
                        )
                        segment.muted = True
                        self._muted_segments.add(segment)

                    else:
                        self._log.error(
                            "Re-downloaded .ts segment %s does not match corrupt one.",
                            segment.id,
                        )

                # occasionally the last few pieces of a stream may not be archived to the VOD and so won't be
                # re-downloaded. instead we just assume the corrupt segment is OK.
                except FileNotFoundError:
                    self._log.debug(
                        "Segment %s could not be re-downloaded - it may no longer be available so the "
                        "potentially corrupt segment will be used.",
                        segment.id,
                    )
                    self._completed_segments.add(segment)
                    shutil.move(
                        Path(self.output_dir, "parts", segment_fp + ".corrupt"),
                        Path(self.output_dir, "parts", segment_fp),
                    )
                    segment.muted = True
                    self._muted_segments.add(segment)

        except CorruptPartError as exc:
            raise VideoDownloadError(
                "Corrupt part(s) still present after retrying VOD download. Ensure VOD is still "
                "available and either delete the listed #####.ts part(s) from 'parts' folder or entire "
                f"'parts' folder if issue persists.\n{str(sorted(corruption))}"
            ) from exc

    def merge(self):
        """
        Attempt to merge downloaded VOD parts and verify them.
        """
        merger = Merger(
            self.vod,
            self.output_dir,
            self._completed_segments,
            self._muted_segments,
            self._quiet,
        )

        # attempt to merge
        try:
            merger.merge()

        # corrupt part(s) encountered - redownload them and reattempt merge
        except CorruptPartError as _c:
            self.repair_vod_corruptions(_c.parts)
            merger.set_muted_segments(self._muted_segments)
            merger.merge()

        except Exception as exc:
            raise VideoMergeError("Exception raised while merging VOD.") from exc

        # verify VOD based on its length
        if merger.verify_length():
            merger.cleanup_temp_files()

        else:
            raise VideoMergeError(
                "VOD verification failed as VOD length is outside the acceptable range."
            )

    def cleanup_temp_files(self):
        """
        Deletes temporary and transitional files used for archiving VOD video.
        """
        self._log.debug("Deleting VOD parts - this may take a while.")
        shutil.rmtree(Path(self.output_dir, "parts"), ignore_errors=True)
        shutil.rmtree(
            Path(tempfile.gettempdir(), "twitch-archiver", str(self.vod.v_id)),
            ignore_errors=True,
        )


class Merger:
    """
    Class used for merging downloaded .ts segments into a single file and performing verification.
    """

    _log = logging.getLogger()
    _api = Api()

    def __init__(self, vod, output_dir, completed_segments, muted_segments, quiet):
        """
        Class constructor.
        """
        self.vod = vod
        self._output_dir = output_dir
        self._completed_segments = completed_segments
        self._completed_parts = self.get_completed_parts()
        self._muted_segment_ids = [s.id for s in muted_segments]
        self._quiet = quiet

    def set_muted_segments(self, segments):
        self._muted_segment_ids = [s.id for s in segments]

    def merge(self):
        """
        Merges downloaded TS segments into a single .mp4

        :raises CorruptPartError: when corruptions are found outside muted segments
        :raises VodConvertError: unrecoverable error ocurred when trying to convert VOD
        """
        # export vod chapters
        self._write_chapters()

        # export vod thumbnail
        if self.vod.thumbnail_url:
            if (
                self.vod.thumbnail_url
                != "https://vod-secure.twitch.tv/_404/404_processing_90x60.png"
            ):
                self._write_thumbnail()

        # merge and remux mpegts segments to single mp4
        self._log.info("Merging VOD parts. This may take a while.")
        self._combine_vod_parts()

        self._log.info("Converting VOD to mp4. This may take a while.")
        self._convert_vod()

    def _write_chapters(self):
        # retrieve vod chapters
        try:
            vod_chapters = self.vod.get_chapters()
            # write chapters to file
            with open(
                Path(self._output_dir, "chapters.json"), "w", encoding="utf8"
            ) as chapters_file:
                chapters_file.write(str(vod_chapters))

            # format and write vod chapters to parts dir
            with open(
                Path(self._output_dir, "parts", "chapters.txt"), "w", encoding="utf8"
            ) as chapters_file:
                chapters_file.write(str(format_vod_chapters(vod_chapters)))

        except Exception as exc:
            self._log.error(
                "Failed to retrieve or insert chapters into VOD file. %s", exc
            )

    def _combine_vod_parts(self):
        """
        Combines the downloaded VOD .ts parts.
        """
        _progress = Progress()

        # concat files if all pieces present, otherwise fall back to using ffmpeg
        _final_part_id = max([_s.id for _s in self._completed_segments])

        _dicontinuity = set([i for i in range(_final_part_id + 1)]).difference(
            [_s.id for _s in self._completed_segments]
        )
        if not _dicontinuity:
            # merge all .ts files by concatenating them
            with open(str(Path(self._output_dir, "merged.ts")), "wb") as _merged_file:
                _pt = 0
                for _part in self._completed_parts:
                    _pt += 1
                    # append part to merged file
                    with open(Path(self._output_dir, "parts", _part), "rb") as _ts_part:
                        shutil.copyfileobj(_ts_part, _merged_file)

                    if not self._quiet:
                        _progress.print_progress(_pt, len(self._completed_segments))

        else:
            # merge all .ts files with ffmpeg concat demuxer as missing segments can cause corruption with
            # other method

            self._log.debug(
                "Discontinuity found, merging with ffmpeg.\n Discontinuity: %s",
                _dicontinuity,
            )

            # create file with list of parts for ffmpeg
            with open(
                Path(self._output_dir, "parts", "segments.txt"), "w", encoding="utf8"
            ) as _segment_file:
                for _part in self._completed_parts:
                    _segment_file.write(
                        f"file '{Path(self._output_dir, 'parts', _part)}'\n"
                    )

            with subprocess.Popen(
                f"ffmpeg -hide_banner -fflags +genpts -f concat -safe 0 -y -i "
                f'"{str(Path(self._output_dir, "parts", "segments.txt"))}" '
                f'-c copy "{str(Path(self._output_dir, "merged.ts"))}"',
                shell=True,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                encoding="cp437",
            ) as _p:
                # get progress from ffmpeg output and print progress bar
                if not self._quiet:
                    for _line in _p.stderr:
                        if "time=" in _line.rstrip():
                            # extract current timestamp from output
                            _cur_time = (
                                re.search("(?<=time=).*(?= bitrate=)", _line)
                                .group(0)
                                .split(":")
                            )
                            _cur_time = (
                                int(_cur_time[0]) * 3600
                                + int(_cur_time[1]) * 60
                                + int(_cur_time[2][:2])
                            )

                            _progress.print_progress(int(_cur_time), self.vod.duration)

                if _p.returncode:
                    self._log.error(
                        "VOD merger exited with error. Command: %s.", _p.args
                    )
                    raise VideoConvertError(
                        f"VOD merger exited with error. Command: {_p.args}."
                    )

    def _convert_vod(self):
        """Converts the VOD from a .ts format to .mp4.

        :raises vodConvertError: error encountered during conversion process
        """
        _progress = Progress()
        _corrupt_parts: set[MpegSegment] = set()

        # get dts offset of first available part
        _dts_offset = self._get_dts_offset()

        # create ffmpeg command
        _ffmpeg_cmd = (
            f'ffmpeg -hide_banner -y -i "{Path(self._output_dir, "merged.ts")}" '
        )
        # insert metadata if present
        if Path(self._output_dir, "parts", "chapters.txt").exists():
            _ffmpeg_cmd += f'-i "{Path(self._output_dir, "parts", "chapters.txt")}" -map_metadata 1 '

        _ffmpeg_cmd += f'-c:a copy -c:v copy "{Path(self._output_dir, "vod.mp4")}"'

        # convert merged .ts file to .mp4
        with subprocess.Popen(
            _ffmpeg_cmd,
            shell=True,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            encoding="cp437",
        ) as _p:
            # get progress from ffmpeg output and catch corrupt segments
            _ffmpeg_log = ""
            for line in _p.stderr:
                _ffmpeg_log += line.rstrip()
                if "time=" in line:
                    # extract current timestamp from output
                    _cur_time = (
                        re.search(r"(?<=time=).*(?= bitrate=)", line)
                        .group(0)
                        .split(":")
                    )
                    _cur_time = (
                        int(_cur_time[0]) * 3600
                        + int(_cur_time[1]) * 60
                        + int(_cur_time[2][:2])
                    )

                    if not self._quiet:
                        _progress.print_progress(int(_cur_time), self.vod.duration)

                elif "Packet corrupt" in line:
                    try:
                        _dts_timestamp = int(
                            re.search(r"(?<=dts = ).*(?=\).)", line).group(0)
                        )

                    # Catch corrupt parts without timestamp, shows up as 'NOPTS'
                    except ValueError as exc:
                        raise VideoConvertError(
                            "Corrupt packet encountered at unknown timestamp while converting VOD. "
                            "Delete 'parts' folder and re-download VOD."
                        ) from exc

                    _corrupt_part = MpegSegment(
                        floor((_dts_timestamp - _dts_offset) / 90000 / 10), 10
                    )

                    # ignore if corrupt packet within ignore_corruptions range
                    if _corrupt_part.id in self._muted_segment_ids:
                        _corrupt_part.muted = True
                        self._log.debug(
                            "Ignoring corrupt packet as part in whitelist. Part: %s",
                            _corrupt_part,
                        )

                    else:
                        _corrupt_parts.add(_corrupt_part)
                        self._log.error(
                            "Corrupt packet encountered. Part: %s", _corrupt_part
                        )

        if _p.returncode:
            self._log.debug(
                "FFmpeg exited with error code, output dumped to VOD directory."
            )
            with open(
                Path(self._output_dir, "parts", "ffmpeg.log"), "w", encoding="utf8"
            ) as _ffout:
                _ffout.write(_ffmpeg_log)

            raise VideoConvertError(
                "VOD converter exited with error. Delete 'parts' directory and re-download VOD."
            )

        if _corrupt_parts:
            # raise error so we can try to recover
            raise CorruptPartError(_corrupt_parts)

    def _get_dts_offset(self):
        """
        Finds the DTS offset for a given stream based on the lowest available part.
        """

        # fetch parts from dir
        _parts = self.get_completed_parts()

        if _parts:
            # fetch id of first part
            _part_id = int(_parts[0].replace(".ts", ""))

            with subprocess.Popen(
                f"ffprobe -v quiet -print_format json -show_format -show_streams "
                f'"{Path(self._output_dir, "parts", _parts[0])}"',
                shell=True,
                stdout=subprocess.PIPE,
                universal_newlines=True,
                encoding="cp437",
            ) as _p:
                _ts_file_data = ""
                for _line in _p.stdout:
                    _ts_file_data += _line.rstrip()

                # retrieve dts start_time of part, subtracting the offset of the part itself and multiplying the value
                # by 90000 (default timescale)
                return (
                    float(json.loads(_ts_file_data)["format"]["start_time"])
                    - (_part_id * 10)
                ) * 90000

        raise FileNotFoundError("No parts available to fetch DTS offset of stream.")

    def verify_length(self):
        """Verifies the length of the downloaded VOD.

        :return: True if VOD length within the acceptable range
        :rtype: bool
        :raises VodVerificationError: if error occurs when verifying
        """
        self._log.debug("Verifying length of VOD file.")

        # skip verification if .ignorelength present
        if Path(self._output_dir, ".ignorelength").is_file():
            self._log.debug(".ignorelength file present - skipping verification.")
            return True

        # retrieve vod file duration
        _p = subprocess.run(
            f'ffprobe -v quiet -i "{Path(self._output_dir, "vod.mp4")}" '
            f"-show_entries format=duration -of default=noprint_wrappers=1:nokey=1",
            shell=True,
            capture_output=True,
            encoding="cp437",
        )

        if _p.returncode:
            raise VideoVerificationError(
                f"VOD length verification exited with error. Command: {_p.args}."
            )

        try:
            downloaded_length = int(float(_p.stdout.rstrip()))

        except Exception as exc:
            raise VideoVerificationError(
                "Failed to fetch downloaded VOD length. See log for details."
            ) from exc

        self._log.debug(
            "Downloaded VOD length is %s. Expected length is %s.",
            downloaded_length,
            self.vod.duration,
        )

        # pass verification if downloaded file is within 2s of expected length
        if 2 >= downloaded_length - self.vod.duration >= -2:
            return True

        return False

    def _write_thumbnail(self):
        """
        Downloads and stores the thumbnail for the VOD.
        """
        try:
            self._log.debug("Downloading VOD thumbnail.")
            thumbnail = self._api.get_request(
                self.vod.thumbnail_url.replace("%{width}x%{height}", "1920x1080")
            )
            with open(Path(self._output_dir, "thumbnail.jpg"), "wb") as thumbnail_file:
                thumbnail_file.write(thumbnail.content)

        except Exception as exc:
            self._log.error("Failed to grab thumbnail for VOD. Error: %s", str(exc))

    def get_completed_parts(self):
        """
        Generates a list of segments based on the completed IDs.

        :return: list of segments padded to 5 digits with .ts extension
        """
        return [f"{seg.id:05d}.ts" for seg in self._completed_segments]

    def cleanup_temp_files(self):
        """
        Deletes temporary and transitional files used for archiving VOD video.
        """
        Path(self._output_dir, "merged.ts").unlink()



import os
import tempfile
from multiprocessing import Queue
from pathlib import Path

from twitcharchiver import Configuration
from twitcharchiver.api import Api
from twitcharchiver.downloader import Downloader
from twitcharchiver.downloaders.chat import Chat
from twitcharchiver.downloaders.stream import Stream
from twitcharchiver.downloaders.video import Video
from twitcharchiver.logger import ProcessWithLogging, ProcessLogger
from twitcharchiver.vod import Vod, ArchivedVod


class RealTime(Downloader):
    """
    Class used for downloading currently live broadcasts, using parallel download functions to grab both the stream
    and VOD video and chat.
    """

    # class vars
    _api: Api = Api()
    _quality: str = ""

    def __init__(
        self,
        vod: Vod,
        parent_dir: Path = os.getcwd(),
        archive_chat: bool = True,
        quality: str = "best",
        threads: int = 20,
    ):
        """
        Class constructor.

        :param vod: VOD to be downloaded
        :type vod: Vod
        :param parent_dir: path to parent directory for downloaded files
        :type parent_dir: str
        :param archive_chat: boolean whether the chat logs should also be grabbed
        :type archive_chat: bool
        :param quality: quality to download in the format [resolution]p[framerate], or either 'best' or 'worst'
        :type quality: str
        :param threads: number of worker threads to use when downloading
        :type threads: int
        """
        super().__init__(parent_dir, True)

        self.vod = vod
        self.parent_dir = parent_dir
        self.archive_chat = archive_chat
        self.quality = quality
        self.threads = threads

        self.chat = None
        self.stream = None
        self.video = None

    def start(self):
        # log files are stored in either the provided log directory or %TEMP%/STREAM_ID
        # we change to this directory as the multiprocessing logger has difficulties with passing
        # variables into it
        logging_dir = Path(tempfile.gettempdir(), str(self.vod.s_id))
        conf = Configuration.get()
        if conf["log_dir"]:
            logging_dir = Path(conf["log_dir"])

        _q = Queue()

        # create downloaders
        self.chat = Chat(self.vod, self.parent_dir, True)
        self.stream = Stream(
            self.vod.channel, self.vod, self.parent_dir, self.quality, True
        )
        self.video = Video(self.vod, self.parent_dir, self.quality, self.threads, True)

        Path(logging_dir).mkdir(exist_ok=True, parents=True)
        # logging directory is used and moved into as Windows doesn't properly share the global logger, so it is
        # reconfigured using a relative path as it is much easier than passing a designated file to the
        # processlogger instance.
        os.chdir(logging_dir)

        process_logger = ProcessLogger.create_global_logger()
        process_logger.start()

        workers = [
            ProcessWithLogging(target=self.stream.start),
            ProcessWithLogging(target=self.video.start, args=[_q]),
        ]

        if self.archive_chat:
            workers.append(ProcessWithLogging(target=self.chat.start))

        try:
            for _w in workers:
                _w.start()

            # get returned video downloader
            self.video: Video = _q.get()

            # wait until all workers are done
            for _w in workers:
                _w.join()

            self._handle_errors(workers)

        finally:
            _q.close()
            _q.join_thread()

            if process_logger:
                process_logger.stop()
                process_logger.join()

    def _handle_errors(self, workers):
        errors = []

        # discover errors based on exit codes of archivers
        # [0] is stream worker
        if workers[0].exitcode == 1:
            self._log.error("Real-time stream archiver exited with error.")
            errors.append(Stream)

        # [1] is video worker
        if workers[1].exitcode == 1:
            self._log.error("Real-time video archiver exited with error.")
            errors.append(Video)

        # [2] is chat worker
        if len(workers) > 2:
            if workers[2].exitcode == 1:
                self._log.error("Real-time chat archiver exited with error.")
                errors.append(Chat)

        # handle various error cases. stream archiver failing is recoverable as long as video archiver finishes
        # successfully.
        if Video in errors:
            self._log.error(
                "Real-time archiver failed as video archiver exited with error. "
                "See log for details."
            )
        else:
            if isinstance(self.vod, ArchivedVod):
                self.vod.video_archived = True

        if Chat in errors:
            self._log.error(
                "Real-time archiver failed as chat archiver exited with error. "
                "See log for details."
            )
        else:
            if self.archive_chat and isinstance(self.vod, ArchivedVod):
                self.vod.chat_archived = True

    def export_metadata(self):
        self.video.export_metadata()

    def merge(self):
        self.video.merge()

    def cleanup_temp_files(self):
        self.chat.cleanup_temp_files()
        self.stream.cleanup_temp_files()
        self.video.cleanup_temp_files()



# reference
#   https://stackoverflow.com/questions/7621897/python-logging-module-globally

"""
Logging class used by Twitch Archiver.
"""
import logging
import logging.handlers
import multiprocessing
import sys
import traceback
from pathlib import Path

CONSOLE_FORMATTER = logging.Formatter(
    "%(asctime)s [%(levelname)8s] %(message)s", "%Y-%m-%d %H:%M:%S"
)
FILE_FORMATTER = logging.Formatter(
    "%(asctime)s [%(filename)s:%(lineno)s - %(funcName)s()] %(message)s",
    "%Y-%m-%d %H:%M:%S",
)


class Logger:
    """
    Sets up logging for the script.
    """

    @staticmethod
    def setup_logger(quiet: bool = False, debug: bool = False, logging_dir: str = ""):
        """Sets up logging module.

        :return: python logging object
        """
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)

        # setup console logging
        console = logging.StreamHandler(sys.stdout)
        console.setFormatter(CONSOLE_FORMATTER)

        # check if stream handler already created - necessary as Windows doesn't properly share the global logger.
        if len(logger.handlers) < 1:
            logger.addHandler(console)

        # setup debugging / quiet / file logging
        if quiet:
            logger = Logger.setup_quiet(logger)
        elif debug:
            logger = Logger.setup_debugging(logger)

        if logging_dir:
            logger = Logger.setup_file(logger, logging_dir)

        Logger.suppress_unnecessary()

        return logger

    @staticmethod
    def setup_file(logger, logging_dir: str):
        """
        Sets up file logging with given logger and filepath.

        :param logger: logging object to enable file output on
        :param logging_dir: directory to store log file(s)
        :return: logging object
        """
        if Path(logging_dir).is_file():
            raise FileExistsError(
                "Error configuring logging, file exists in place of log directory."
            )

        Path(logging_dir).mkdir(parents=True, exist_ok=True)

        # use rotating file handler with max size of 100MB * 5
        file = logging.handlers.RotatingFileHandler(
            Path(logging_dir, "debug.log"),
            maxBytes=100000000,
            backupCount=5,
            encoding="utf8",
        )
        file.setLevel(logging.DEBUG)
        file.setFormatter(FILE_FORMATTER)
        logger.addHandler(file)

        return logger

    @staticmethod
    def setup_debugging(logger):
        logger.setLevel(logging.DEBUG)

        return logger

    @staticmethod
    def setup_quiet(logger):
        logger.setLevel(50)

        return logger

    @staticmethod
    def suppress_unnecessary():
        # supress other messages
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("charset_normalizer").setLevel(logging.WARNING)


# logging handler used with realtime archiver
# reference:
#   https://stackoverflow.com/a/60831737
class ProcessLogger(multiprocessing.Process):
    _global_process_logger = None

    def __init__(self):
        super().__init__()
        self.queue = multiprocessing.Queue(-1)

    @classmethod
    def get_global_logger(cls):
        if cls._global_process_logger is not None:
            return cls._global_process_logger
        raise Exception("No global process logger exists.")

    @classmethod
    def create_global_logger(cls):
        cls._global_process_logger = ProcessLogger()
        return cls._global_process_logger

    @staticmethod
    def configure():
        root = Logger.setup_logger()
        # The global logging instance isn't properly shared with `spawn` and so it has to be re-added when we setup
        # multiprocess logging.
        # limit to 5x 100MB log files
        h = logging.handlers.RotatingFileHandler(
            "debug.log", "a", 100 * 1024**2, 5, encoding="utf8"
        )
        h.setFormatter(FILE_FORMATTER)
        root.addHandler(h)

    def stop(self):
        self.queue.put_nowait(None)

    def run(self):
        self.configure()
        while True:
            try:
                record = self.queue.get()
                if record is None:
                    break
                logger = logging.getLogger(record.name)
                logger.handle(record)
            except Exception:
                print("Multiprocess logger error:", file=sys.stderr)
                traceback.print_exc(file=sys.stderr)


def configure_new_process(log_process_queue):
    h = logging.handlers.QueueHandler(log_process_queue)
    root = logging.getLogger()
    root.addHandler(h)
    # grab level from main logger
    root.setLevel(root.handlers[0].level)
    Logger.suppress_unnecessary()


class ProcessWithLogging(multiprocessing.Process):
    def __init__(self, target, args=None, kwargs=None, log_process=None):
        super().__init__()
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        self.target = target
        self.args = args
        self.kwargs = kwargs

        if log_process is None:
            log_process = ProcessLogger.get_global_logger()
        self.log_process_queue = log_process.queue

    def run(self):
        configure_new_process(self.log_process_queue)
        self.target(*self.args, **self.kwargs)



"""
Class for retrieving and storing information related to channels and users.
"""
import logging
from datetime import datetime, timezone
from random import randrange

import m3u8

from twitcharchiver.api import Api
from twitcharchiver.utils import time_since_date


class Channel:
    """
    Class for storing and retrieving data for a single Twitch channel / user.
    """

    def __init__(self, channel_name: str = "", channel_id: int = 0, owner: dict = None):
        """
        Initialize class variables.

        :param channel_name: name of channel
        :type channel_name: str
        :param owner: dict of channel values retrieved from Twitch
        :type owner: dict
        """
        self._api: Api = Api()
        self._log = logging.getLogger()

        self.id: int = channel_id
        self.name: str = channel_name
        self.display_name = ""
        self.stream: dict = {}
        self._broadcast_v_id = int()

        self._last_update: float = 0

        if owner:
            self._parse_dict(owner)

        elif self.name or self.id:
            self._parse_dict(self._fetch_metadata())

    def __repr__(self):
        return str(self.get_info())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return bool(self.id == other.id)
        raise TypeError

    def __hash__(self):
        return hash(self.id)

    def __bool__(self):
        return bool(self.id)

    def get_info(self):
        """
        Returns a dict of important channel information.

        :return: dict of channel id, name and stream info.
        :rtype: dict
        """
        return {
            "id": self.id,
            "name": self.name,
            "display_name": self.display_name,
            "stream": self.stream,
        }

    def _parse_dict(self, owner: dict):
        """
        Parses information from 'owner' object returned from Twitch.

        :param owner: channel owner object retrieved from Twitch
        """
        self.id = int(owner["id"])
        self.name = owner["login"]
        self.display_name = owner["displayName"]
        self.stream = owner["stream"]

    def _fetch_metadata(self):
        """
        Fetches metadata from Twitch regarding the channel.

        :return: retrieved user data
        :rtype dict
        """
        if self.id and not self.name:
            self.name = self._user_from_id(self.id)["login"]

        _r = self._api.gql_request(
            "ChannelShell",
            "580ab410bcd0c1ad194224957ae2241e5d252b2c5173d8e0cce9d32d5bb14efe",
            {"login": f"{self.name}"},
        )
        _user_data = _r.json()[0]["data"]["userOrError"]
        self._log.debug("User data for %s: %s", self.name, _user_data)

        # failure return contains "userDoesNotExist" key
        if "userDoesNotExist" not in _user_data.keys():
            self._last_update = datetime.now(timezone.utc).timestamp()
            return _user_data

        return {}

    def _user_from_id(self, channel_id: int):
        """
        Fetch information for the channel with the provided ID.

        :param channel_id: id of channel to fetch
        :return: dict of channel information
        :rtype: dict
        """
        _r = self._api.get_request(
            f"https://api.twitch.tv/helix/users?id={channel_id}",
            h={"Client-ID": "gh70y1spw727ohtgzbhc0hppvq9br2"},
        )

        return _r.json()["data"][0]

    def is_live(self, force_refresh=False):
        """
        Checks if the channel is currently live.

        :param force_refresh: True if cache to be ignored and metadata to be forcefully refreshed
        :type force_refresh: bool
        :return: True if channel live
        :rtype: bool
        """
        # refresh metadata if it was last updated more than 60 seconds ago
        if time_since_date(self._last_update) > 60 or force_refresh:
            self.refresh_metadata()
        return self.stream is not None

    def refresh_metadata(self):
        """
        Refreshes all metadata for the channel.
        """
        self._parse_dict(self._fetch_metadata())

    def get_stream_info(self):
        """Retrieves information relating to a channel if it is currently live.

        :return: dictionary with information about a channel's live stream if any
        :rtype: dict
        """
        _query_vars = {
            "channel": self.name,
            "clipSlug": "",
            "isClip": False,
            "isLive": True,
            "isVodOrCollection": False,
            "vodID": "",
        }

        _r = self._api.gql_request(
            "ComscoreStreamingQuery",
            "e1edae8122517d013405f237ffcc124515dc6ded82480a88daef69c83b53ac01",
            _query_vars,
        )

        _stream_info = _r.json()[0]["data"]["user"]

        if _stream_info:
            self._log.debug("Stream info for %s: %s", self.name, _stream_info)
            return _stream_info

        self._log.debug("No broadcast info found for %s", self.name)
        return {
            "id": "",
            "displayName": "",
            "stream": {},
            "broadcastSettings": {
                "id": "",
                "title": "",
                "__typename": "BroadcastSettings",
            },
            "__typename": "User",
        }

    def _get_broadcast_v_id(self):
        """
        Fetches the paired VOD ID for the currently live broadcast.

        :return: VOD ID
        :rtype: int
        """
        _r = self._api.gql_request(
            "ChannelVideoLength",
            "ac644fafd686f2cb0e3864075af7cf3bb33f4e0525bf84921b10eabaa4e048b5",
            {"channelLogin": f"{self.name.lower()}"},
        )
        _channel_video_length = _r.json()[0]["data"]["user"]["videos"]["edges"]

        if _channel_video_length:
            broadcast_info = _channel_video_length[0]["node"]["id"]
            self._log.debug("Live broadcast info: %s", broadcast_info)
            return int(broadcast_info)

        self._log.debug("No data returned by ChannelVideoLength API for %s.", self.name)
        return int()

    @property
    def broadcast_v_id(self):
        """
        Fetch, store and return most recent broadcast VOD ID.
        """
        if not self._broadcast_v_id:
            self._broadcast_v_id = self._get_broadcast_v_id()

        return self._broadcast_v_id

    def get_stream_index(self, quality: str = "best"):
        """
        Retrieves a m3u8 index of the channel's live stream.

        :param quality: desired quality in the format [resolution]p[framerate] or 'best', 'worst'
        :type quality: str
        :return: url of m3u8 playlist
        :rtype: str
        """
        from twitcharchiver.vod import Vod

        _access_token = self.get_stream_playback_access_token()

        _p = {
            "player": "twitchweb",
            "fast_bread": "true",
            "token": _access_token["value"],
            "sig": _access_token["signature"],
            "allow_source": "true",
            "playlist_include_framerate": "true",
            "player_backend": "mediaplayer",
            "supported_codecs": "avc1",
            "p": randrange(1000000, 9999999),
        }
        _r = self._api.get_request(
            f"https://usher.ttvnw.net/api/channel/hls/{self.name.lower()}.m3u8", p=_p
        )

        _index = m3u8.loads(_r.text)

        # grab 'name' of m3u8 streams - contains [resolution]p[framerate]
        _available_resolutions = [
            m[0].group_id.split("p")
            for m in [m.media for m in _index.playlists]
            if m[0].group_id != "chunked"
        ]
        # insert 'chunked' stream separately as its named differently and strip ' (source)' from name
        _available_resolutions.insert(
            0, _index.media[0].name.strip(" (source)").split("p")
        )
        self._log.debug(
            "Available resolutions for %s are: %s", self.name, _available_resolutions
        )

        _index_url = _index.playlists[
            Vod.get_quality_index(quality, _available_resolutions)
        ].uri
        self._log.debug("Index for broadcast by %s: %s", self.name, _index_url)

        return _index_url

    def get_stream_playlist(self, index_url: str = ""):
        """
        Retrieves the playlist for a given VOD index along with updating the VOD duration.
        """
        if not index_url:
            index_url = self.get_stream_index()

        _stream_playlist = self._api.get_request(index_url).text

        return _stream_playlist

    def get_stream_playback_access_token(self):
        """
        Gets a stream access token for the channel.

        :return: dictionary of playback access token values
        :rtype: dict
        """
        # only accepts the default client ID for non-authenticated clients
        _h = {"Client-Id": "ue6666qo983tsx6so1t0vnawi233wa"}
        _q = """
        {{
            streamPlaybackAccessToken(
                channelName: "{channel}",
                params: {{
                    platform: "web",
                    playerBackend: "mediaplayer",
                    playerType: "embed"
                }}
            ) {{
                signature
                value
            }}
        }}
        """.format(
            channel=self.name.lower()
        )
        _r = self._api.post_request("https://gql.twitch.tv/gql", j={"query": _q}, h=_h)

        _access_token = _r.json()["data"]["streamPlaybackAccessToken"]
        self._log.debug("Access token retrieved for %s. %s", self.name, _access_token)

        return _access_token

    def _get_most_recent_videos(self):
        """
        Retrieves all recent broadcasts for the channel.

        :return: list of Vod objects
        :rtype: list[Vod]
        """
        from twitcharchiver.vod import Vod

        _r = self._api.gql_request(
            "FilterableVideoTower_Videos",
            "a937f1d22e269e39a03b509f65a7490f9fc247d7f83d6ac1421523e3b68042cb",
            {
                "broadcastType": "ARCHIVE",
                "channelOwnerLogin": f"{self.name.lower()}",
                "limit": 30,
                "videoSort": "TIME",
            },
        )

        _recent_videos = [
            Vod(vod_info=v["node"])
            for v in _r.json()[0]["data"]["user"]["videos"]["edges"]
        ]

        self._log.debug("Recent videos for %s: %s", self.name, _recent_videos)
        return _recent_videos

    def get_latest_video(self):
        """Retrieves the latest VOD for the channel.

        :return: The most recent VOD for the channel
        :rtype: Vod
        """
        return self._get_most_recent_videos()[0]

    def get_channel_videos(self):
        """
        Retrieves all available VODs for the channel.

        :return: list of all available VODs for the channel
        :rtype: list[Vod]
        """
        from twitcharchiver.vod import Vod

        _channel_videos = []
        _query_vars = {
            "broadcastType": "ARCHIVE",
            "channelOwnerLogin": f"{self.name.lower()}",
            "limit": 30,
            "videoSort": "TIME",
        }

        while True:
            _r = self._api.gql_request(
                "FilterableVideoTower_Videos",
                "a937f1d22e269e39a03b509f65a7490f9fc247d7f83d6ac1421523e3b68042cb",
                _query_vars,
            )

            # retrieve list of videos from response
            _videos = [
                Vod(vod_info=v["node"])
                for v in _r.json()[0]["data"]["user"]["videos"]["edges"]
            ]
            _channel_videos.extend(_videos)

            if (
                _r.json()[0]["data"]["user"]["videos"]["pageInfo"]["hasNextPage"]
                is not False
            ):
                # set cursor
                _query_vars["cursor"] = _r.json()[0]["data"]["user"]["videos"]["edges"][
                    -1
                ]["cursor"]

            else:
                break

        if _channel_videos:
            self._log.debug(
                "VODs retrieved for %s: %s", self.name, len(_channel_videos)
            )
            return _channel_videos

        self._log.debug("No VODs found for %s.", self.name)
        return []



"""
Handles communication with the Twitch API.
"""
import logging
from time import sleep

import requests

from twitcharchiver.exceptions import (
    RequestError,
    TwitchAPIError,
    TwitchAPIErrorNotFound,
    TwitchAPIErrorForbidden,
    TwitchAPIErrorBadRequest,
)


class Api:
    """
    Handles communication with the Twitch API.
    """

    def __init__(self):
        """
        Class constructor.
        """
        self._session = requests.session()
        self._headers = {}
        self.logging = logging.getLogger()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # cleanup
        self.close()

    def close(self):
        """
        Cleanly shutdown requests session.
        """
        self._session.close()

    def add_headers(self, headers: dict):
        """
        Adds desired headers to all calls.

        :param headers: dictionary of header values to add
        :return: updated headers
        :rtype: dict
        """
        self._headers.update(headers)
        return self._headers

    def get_request(self, url: str, p: dict = None, h: dict = None):
        """
        Wrapper for get requests for catching exceptions and status code issues.\n

        :param url: http/s endpoint to send request to
        :param p: parameter(s) to pass with request
        :param h: header to pass with request (overrides class headers)
        :return: entire requests response
        :raises requestError: on requests module error
        :raises TwitchAPIErrorBadRequest: on http code 400
        :raises TwitchAPIErrorForbidden: on http code 403
        :raises TwitchAPIErrorNotFound: on http code 404
        :raises TwitchAPIError: on any http code other than 400, 403, 404 or 200
        """
        _headers = h or self._headers
        _params = p

        try:
            _r = self._session.get(url, headers=_headers, params=_params, timeout=10)

        except requests.exceptions.RequestException as err:
            raise RequestError(url, err) from err

        if _r.status_code == 400:
            raise TwitchAPIErrorBadRequest(_r)

        if _r.status_code == 403:
            raise TwitchAPIErrorForbidden(_r)

        if _r.status_code == 404:
            raise TwitchAPIErrorNotFound(_r)

        if _r.status_code != 200:
            raise TwitchAPIError(_r)

        return _r

    def post_request(self, url, d=None, j=None, h=None):
        """
        Wrapper for post requests for catching exceptions and status code issues.

        :param url: http/s endpoint to send request to
        :type url: str
        :param d: formatted data to send with request
        :type d: str
        :param j: json to send with request
        :type j: dict or list
        :param h: override class headers to send with request
        :type h: dict
        :return: entire requests response
        """
        if not (d or j):
            raise ValueError(
                "Either data (d) or json (j) must be included with request."
            )

        try:
            if j is None:
                _r = self._session.post(
                    url, data=d, headers=h if h else self._headers, timeout=10
                )

            elif d is None:
                _r = self._session.post(
                    url, json=j, headers=h if h else self._headers, timeout=10
                )

            if _r.status_code != 200:
                raise TwitchAPIError(_r)

            return _r

        except requests.exceptions.RequestException as err:
            raise RequestError(url, err) from err

    def gql_request(self, operation: str, query_hash: str, variables: dict):
        """
        Post a gql query and returns the response.

        :param operation: name of operation
        :param query_hash: hash of operation
        :param variables: dict of variable to post with request
        :return: entire request response
        :rtype: requests.Response
        """
        # Uses default client header
        _h = {"Client-Id": "ue6666qo983tsx6so1t0vnawi233wa"}
        _q = [
            {
                "extensions": {
                    "persistedQuery": {"sha256Hash": query_hash, "version": 1}
                },
                "operationName": operation,
                "variables": variables,
            }
        ]

        # retry loop for 'service error' responses
        _attempt = 0
        while True:
            _r = self.post_request("https://gql.twitch.tv/gql", j=_q, h=_h)

            if _attempt >= 5:
                self.logging.error(
                    "Maximum attempts reached while querying GQL API. Error: %s",
                    _r.json(),
                )
                raise TwitchAPIError(_r)

            if "errors" in _r.json()[0].keys():
                _attempt += 1
                self.logging.error(
                    "Error returned when querying GQL API, retrying. Error: %s",
                    _r.json(),
                )
                sleep(_attempt * 10)
                continue

            break

        return _r



"""
Handles parsing and storage of arguments passed to twitch-archiver.
"""
import logging
import re
import sys

from pathlib import Path


class Arguments:
    """
    Function for parsing arguments for access later.
    """

    __args = {}
    _log = logging.getLogger()

    @classmethod
    def setup_args(cls, args):
        """
        Sets up and formats provided arguments.

        :param args: arguments object to create parameters from
        """
        for argument in args:
            cls.set(argument, args[argument])

        if cls.get("show_config"):
            try:
                with open(
                    Path(cls.get("config_dir"), "config.ini"), "r", encoding="utf8"
                ) as config:
                    print(config.read())
                    sys.exit(0)
            except FileNotFoundError:
                sys.exit(
                    "Config file not found. Run Twitch-Archiver once to generate one."
                )

        # validate mutual exclusivity of arguments passed via CLI and environment variables
        # required as values set via environment variables bypass argparse mutex handling
        for mutex_args in (("vod", "channel"), ("live_only", "archive_only")):
            mutex_arg_0, mutex_arg_1 = cls.get(mutex_args[0]), cls.get(mutex_args[1])
            # check if both mutex args have been set
            if mutex_arg_0 and mutex_arg_1:
                raise ValueError(
                    "Cannot accept both of the following mutually exclusive arguments: '"
                    f"{mutex_args[0]}={mutex_arg_0}' and '{mutex_args[1]}={mutex_arg_1}'"
                )

        # get both video and chat logs if neither selected
        if not cls.get("chat") and not cls.get("video"):
            cls.set("chat", True)
            cls.set("video", True)

        # generate list from comma-separated VODs
        try:
            if cls.get("vod"):
                cls.extract_vods_and_channels("vod")

            # generate list from comma-separated channels
            elif cls.get("channel"):
                cls.extract_vods_and_channels("channel")
        except TypeError:
            print("Error parsing provided channel or VOD argument.")

        # split quality into [resolution, framerate]
        if cls.get("quality") not in ["best", "worst"]:
            cls.set("quality", cls.get("quality").split("p"))

        if cls.get("watch"):
            print("Launching Twitch-Archiver in watch mode.")

    @classmethod
    def set(cls, name, value):
        """
        Set a specified class attribute.

        :param name: name of attribute to change
        :param value: value to set attribute to
        """
        cls.__args[name] = value

    @classmethod
    def get(cls, name=None):
        """
        Retrieve a specified attribute.

        :param name: name of attribute to retrieve value of or none to return all
        :return: value of requested attribute, or all attributes if none provided
        """
        if name is None:
            return cls.__args

        return cls.__args[name]

    @classmethod
    def extract_vods_and_channels(cls, arg_name: str):
        """
        Parses the provided `-c | --channel` or `-v | --vod-id` argument and interprets the value.

        :param arg_name: either 'channel' or 'vod' to denote which argument to parse
        :type arg_name: str
        :return: list of channels or VOD IDs
        :rtype list
        """
        # separate comma-separated values
        parsed_args: list = []

        # extract vods ids if file being passed
        if cls.get("file"):
            collected: list = []
            for arg in cls.get(arg_name).split(","):
                # convert list to Path() variables and store for further processing
                collected.extend(Arguments.load_file_line_by_line(Path(arg)))

            cls.set(arg_name, ",".join(collected))

        # format urls to just vod ids or channel names
        for arg in cls.get(arg_name).split(","):
            # skip empty args
            if arg == "":
                continue

            # extract VOD ID or channel name if url passed
            if "/videos/" in arg:
                match = re.findall(r"(?<=twitch\.tv/videos/)[0-9]*", arg)
            else:
                match = re.findall(r"(?<=twitch\.tv/)[a-zA-Z]*", arg)

            # store the extracted value or simply pass to passed args if no match found
            if match:
                parsed_args.append(match[0])

            else:
                parsed_args.append(arg)

        cls.set(arg_name, parsed_args)

    @classmethod
    def load_file_line_by_line(cls, file_path: Path):
        """
        Loads a given file line by line into the provided configuration variable.

        :return: list containing all file lines
        :rtype: list[str]
        """
        try:
            with open(Path(file_path), "r") as _fp:
                return [line.rstrip() for line in _fp]

        except Exception as exc:
            cls._log.error("Failed to read from provided input file. %s", exc)
            return None



r"""
 ______ __    __ ____ ______   __ __ __       ____ ____     __ __ __ ____ __ __   ___ ____
|      |  T__T  l    |      T /  |  T  T     /    |    \   /  |  T  l    |  T  | /  _|    \
|      |  |  |  ||  T|      |/  /|  l  |    Y  o  |  D  ) /  /|  l  ||  T|  |  |/  [_|  D  )
l_j  l_|  |  |  ||  |l_j  l_/  / |  _  |    |     |    / /  / |  _  ||  ||  |  Y    _|    /
  |  | l  `  '  !|  |  |  |/   \_|  |  |    |  _  |    \/   \_|  |  ||  |l  :  |   [_|    \
  |  |  \      / j  l  |  |\     |  |  |    |  |  |  .  \     |  |  |j  l \   /|     |  .  Y
  l__j   \_/\_/ |____j l__j \____l__j__j    l__j__l__j\_j\____l__j__|____j \_/ l_____l__j\_j

Created by:
    https://github.com/Brisppy
Inspired by:
    https://github.com/PetterKraabol/Twitch-Chat-Downloader/
    https://github.com/ihabunek/twitch-dl
    https://github.com/streamlink/streamlink
Twitch Archiver - A simple, fast, platform-independent tool for downloading Twitch streams, videos, and chat logs.
Copyright (C) 2023 https://github.com/Brisppy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.
You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import multiprocessing
import os
import sys
import tempfile
import textwrap
from pathlib import Path
from time import sleep

from twitcharchiver.arguments import Arguments
from twitcharchiver.channel import Channel
from twitcharchiver.configuration import Configuration
from twitcharchiver.logger import Logger
from twitcharchiver.processing import Processing
from twitcharchiver.utils import getenv, check_update_available, get_latest_version

__version__ = "4.0.3"

from twitcharchiver.vod import Vod, ArchivedVod


def main():
    """
    Main processing for twitch-archiver.
    """
    parser = argparse.ArgumentParser(
        argument_default=None,
        description=textwrap.dedent(
            f"""\
    Twitch Archiver v{__version__} - https://github.com/Brisppy/twitch-archiver

    A fast, platform-independent Python script for downloading past and present Twitch VODs and chat logs.

    requires one of:
        -c CHANNEL, --channel CHANNEL / URL
                Channel(s) to download, separated with commas or a file path with `-f | --file` arg.
        -v VOD, --vod VOD_ID / URL
                VOD ID(s) to download, separated with commas or a file path with `-f | --file` arg.
                
    Both the video and chat logs are grabbed if neither are specified.
    """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    mode = parser.add_mutually_exclusive_group(
        required=not (
            ("--show-config" in sys.argv)
            or ((getenv("TWITCH_ARCHIVER_CHANNEL")) is not None)
            or (getenv("TWITCH_ARCHIVER_VOD") is not None)
        )
    )
    stream = parser.add_mutually_exclusive_group(required=False)
    loglevel = parser.add_mutually_exclusive_group(required=False)
    mode.add_argument(
        "-c",
        "--channel",
        type=str,
        action="store",
        help="A single twitch channel to download, or multiple comma-separated channels.",
        default=getenv("TWITCH_ARCHIVER_CHANNEL"),
    )
    mode.add_argument(
        "-v",
        "--vod",
        "--vod-id",
        type=str,
        action="store",
        help="A single VOD (e.g 12763849) or many comma-separated IDs (e.g 12763159,12753056).",
        default=getenv("TWITCH_ARCHIVER_VOD_ID"),
    )
    parser.add_argument(
        "-f",
        "--file",
        action="store_true",
        help="Denotes that the value provided to `-c | --channel` or `-v | --vod` is a\n"
        "path to a file. Each line should contain a VOD ID or channel name.",
        default=False,
    )
    parser.add_argument(
        "-C",
        "--chat",
        action="store_true",
        help="Only save chat logs.",
        default=getenv("TWITCH_ARCHIVER_CHAT", False, True),
    )
    parser.add_argument(
        "-V",
        "--video",
        action="store_true",
        help="Only save video.",
        default=getenv("TWITCH_ARCHIVER_VIDEO", False, True),
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        action="store",
        help="Number of video download threads. (default: %(default)s)",
        default=getenv("TWITCH_ARCHIVER_THREADS", 20),
    )
    parser.add_argument(
        "-q",
        "--quality",
        type=str,
        action="store",
        help="Quality to download. Options are 'best', 'worst' or a custom value.\n"
        "Format for custom values is [resolution]p[framerate], (e.g 1080p60, 720p30).\n"
        "(default: best)",
        default="best",
    )
    parser.add_argument(
        "-d",
        "--directory",
        action="store",
        help="Directory to store archived VOD(s), use TWO slashes for Windows paths.\n"
        "(default: current directory)",
        type=Path,
        default=getenv("TWITCH_ARCHIVER_DIRECTORY", Path(os.getcwd())),
    )
    parser.add_argument(
        "-w",
        "--watch",
        action="store_true",
        help="Continually check every 10 seconds for new streams/VODs from a specified channel.",
        default=getenv("TWITCH_ARCHIVER_WATCH", False, True),
    )
    stream.add_argument(
        "-l",
        "--live-only",
        action="store_true",
        default=getenv("TWITCH_ARCHIVER_LIVE_ONLY", False, True),
        help="Only download streams / VODs which are currently live.",
    )
    stream.add_argument(
        "-a",
        "--archive-only",
        action="store_true",
        help="Don't download streams / VODs which are currently live.",
        default=getenv("TWITCH_ARCHIVER_ARCHIVE_ONLY", False, True),
    )
    parser.add_argument(
        "-R",
        "--real-time-archiver",
        action="store_true",
        help="Enable real-time stream archiver.\n"
        "Read https://github.com/Brisppy/twitch-archiver/wiki/Wiki#real-time-archiver.",
        default=getenv("TWITCH_ARCHIVER_REAL_TIME_ARCHIVER", False, True),
    )
    parser.add_argument(
        "-L",
        "--log-dir",
        action="store",
        help="Output logs to specified directory.",
        type=Path,
        default=getenv("TWITCH_ARCHIVER_LOG_FILE", False),
    )
    parser.add_argument(
        "-I",
        "--config-dir",
        action="store",
        type=Path,
        help="Directory to store configuration and VOD database.\n(default: %(default)s)",
        default=getenv(
            "TWITCH_ARCHIVER_CONFIG_DIR",
            Path(os.path.expanduser("~"), ".config", "twitch-archiver"),
        ),
    )
    parser.add_argument(
        "-p",
        "--pushbullet-key",
        action="store",
        help="Pushbullet key for sending pushes on error. Enabled by supplying key.",
        default=getenv("TWITCH_ARCHIVER_PUSHBULLET_KEY", default_val=""),
    )
    loglevel.add_argument(
        "-Q", "--quiet", action="store_true", help="Disable all log output."
    )
    loglevel.add_argument(
        "-D", "--debug", action="store_true", help="Enable debug logs."
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"Twitch Archiver v{__version__}",
        help="Show version number and exit.",
    )
    parser.add_argument(
        "--show-config",
        action="store_true",
        help="Show saved config and exit.",
        default=False,
    )

    # set multiprocessing start mode
    multiprocessing.set_start_method("spawn")

    # setup arguments
    args = Arguments()
    args.setup_args(parser.parse_args().__dict__)

    # setup logging
    log = Logger.setup_logger(args.get("quiet"), args.get("debug"), args.get("log_dir"))
    log.debug("Debug logging enabled.")

    # debug only: output sanitized version of arguments
    args_sanitized = args.get().copy()
    for key in ["pushbullet_key"]:
        if args_sanitized[key]:
            args_sanitized.update({key: 24 * "*" + args_sanitized[key][24:]})

    log.debug("Arguments: %s", args_sanitized)

    # compare with current git version
    latest_version, release_notes = get_latest_version()
    if check_update_available(__version__, latest_version):
        log.warning(
            "New version of Twitch-Archiver available - Version %s:\n"
            "https://github.com/Brisppy/twitch-archiver/releases/latest\nRelease notes:\n\n%s\n",
            latest_version,
            release_notes,
        )
    else:
        log.info(
            "Twitch Archiver v%s - https://github.com/Brisppy/twitch-archiver",
            __version__,
        )

    # load configuration from ini
    config = Configuration()
    config.load_config(Path(args.get("config_dir"), "config.ini"))
    log.debug("Settings prior to loading config: %s", config.get_sanitized())

    # overwrite different or missing configuration variables
    config.generate_config(args.get())
    log.debug("Settings after loading config: %s", config.get_sanitized())

    # create temp dir for downloads and lock files
    Path(tempfile.gettempdir(), "twitch-archiver").mkdir(exist_ok=True)

    process = Processing(Configuration.get())

    if args.get("channel") is not None:
        channels = [Channel(c) for c in args.get("channel")]

        while True:
            process.get_channel(channels)

            if args.get("watch"):
                sleep(10)

            else:
                break

    elif args.get("vod") is not None:
        vods = [ArchivedVod.convert_from_vod(Vod(v)) for v in args.get("vod")]

        process.vod_downloader(vods)


if __name__ == "__main__":
    main()



"""
Module for downloading currently live Twitch broadcasts.
"""
import os
import shutil
import tempfile
from datetime import datetime, timezone
from math import floor
from operator import attrgetter
from pathlib import Path
from time import sleep

import m3u8
import requests

from twitcharchiver.channel import Channel
from twitcharchiver.downloader import Downloader
from twitcharchiver.downloaders.video import MpegSegment, Merger
from twitcharchiver.exceptions import (
    TwitchAPIErrorNotFound,
    UnsupportedStreamPartDuration,
    StreamSegmentDownloadError,
    StreamFetchError,
    StreamOfflineError,
    VideoMergeError,
    RequestError,
)
from twitcharchiver.utils import (
    time_since_date,
    safe_move,
    build_output_dir_name,
    write_json_file,
)
from twitcharchiver.vod import Vod, ArchivedVod

TEMP_BUFFER_LEN = 120
CHECK_INTERVAL = 4


class StreamSegmentList:
    """
    Parses and stores segments of a Twitch livestream and the parts they are derived from.
    """

    def __init__(
        self, stream_created_at: float, align_segments: bool = True, start_id: int = 0
    ):
        self.segments: dict[int:StreamSegment] = {}
        self.current_id = start_id + 1
        self._align_segments = align_segments
        self.stream_created_at = stream_created_at

    def add_part(self, part):
        """
        Adds a given part to the appropriate StreamSegment.

        :param part: part to add
        :type part: StreamSegment.Part
        """
        # generate part id from timestamp if we are aligning segments
        if self._align_segments:
            _parent_segment_id = self._get_id_for_part(part)

        # otherwise use our current id
        else:
            _parent_segment_id = self.current_id

        # create parent segment if one doesn't yet exist
        if _parent_segment_id not in self.segments.keys():
            self.current_id = _parent_segment_id
            self.segments[_parent_segment_id] = StreamSegment(_parent_segment_id)

        # create new segment and increment ID if current segment complete
        if len(self.get_segment_by_id(_parent_segment_id).parts) == 5:
            _parent_segment_id += 1
            self.current_id = _parent_segment_id
            self.segments[_parent_segment_id] = StreamSegment(_parent_segment_id)

        # append part to parent segment
        self.segments[_parent_segment_id].add_part(part)

    def _get_id_for_part(self, part):
        """
        Retrieves the ID for a given part based on it's and the stream's timestamps.

        :param part: Part to retrieve ID for.
        :type part: StreamSegment.Part
        :return: ID of part
        :rtype: int
        """
        # maths for determining the id of a given part based on its timestamp and the stream creation time
        return floor((4 + (part.timestamp - self.stream_created_at)) / 10)

    def is_segment_present(self, segment_id: int):
        """
        Checks if a stream segment ID exists in the StreamSegmentList.

        :param segment_id: ID of segment to look for
        :return: True if segment ID present
        """
        return segment_id in self.segments.keys()

    def get_segment_by_id(self, segment_id: int):
        """
        Fetches the segment with the provided ID.

        :param segment_id: segment id to fetch
        :return: Segment which matches provided ID
        :rtype: StreamSegment
        """
        return self.segments[segment_id]

    def get_completed_segment_ids(self):
        """
        Gathers and returns the ids of all segments with 5 parts.

        :return: set[int]
        """
        _segment_ids: set[int] = set()
        for _segment in self.segments:
            if len(self.segments[_segment].parts) == 5:
                _segment_ids.add(self.segments[_segment].id)

        return _segment_ids

    def pop_segment(self, seg_id):
        """
        Pops the provided segment ID off of the list of segments.

        :param seg_id: id of segment to remove and return
        :return: segment which matches the id
        :rtype: StreamSegment
        """
        return self.segments.pop(seg_id)


class StreamSegment:
    """
    Defines a video segment made up of 5 'Parts', these parts are retrieved from the Twitch API during a livestream.
    In *most* circumstances, each part is 2 seconds long, and can be combined into a 10s 'Segment'. These segments can
    then be matched directly with Segments which make up a Twitch VOD.
    """

    def __init__(self, segment_id: int):
        """
        Defines a video segment made up of 5 StreamSegment parts.

        :param segment_id: ID of segment
        """
        self.parts: list[StreamSegment.Part] = []
        self.id: int = segment_id
        self.duration: float = 0

    class Part:
        """
        A Part advertised by Twitch and later combined into a segment.
        """

        def __init__(self, part):
            """
            Defines a part of a segment.
            """
            self.url: str = part.uri
            self.timestamp: float = part.program_date_time.replace(
                tzinfo=timezone.utc
            ).timestamp()
            self.duration: float = part.duration
            self.title = part.title

        def __repr__(self):
            return str(
                {
                    "url": self.url,
                    "timestamp": self.timestamp,
                    "duration": self.duration,
                }
            )

        def __eq__(self, other):
            if isinstance(other, StreamSegment.Part):
                return self.url == other.url
            raise TypeError

        def __hash__(self):
            return hash(self.url)

    def add_part(self, part: Part):
        """
        Adds a part to the segment and updates the duration.

        :param part: part to add to segment
        """
        self.parts.append(part)
        self.duration += part.duration

    def is_full(self):
        """
        Checks if the segment had five parts.

        :return: True if segment is complete (contains 5 parts)
        :rtype: bool
        """
        if len(self.parts) == 5:
            return True

        return False

    def __repr__(self):
        return str({"id": self.id, "duration": self.duration, "parts": self.parts})

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.parts == other.parts
        raise TypeError

    def __hash__(self):
        return hash(str(self.parts))


class Stream(Downloader):
    """
    Class which handles the downloading of video files for a given Twitch stream.
    """

    _quality: str = ""

    def __init__(
        self,
        channel: Channel,
        vod: Vod = Vod(),
        parent_dir: Path = Path(os.getcwd()),
        quality: str = "best",
        quiet: bool = "False",
    ):
        """
        Class constructor.

        :param channel: Channel to be downloaded
        :type channel: Channel
        :param vod: Optionally provided VOD if trying to sync VOD and stream segments and timestamps
        :type vod: Vod
        :param parent_dir: Parent directory in which to create VOD directory and download to
        :type parent_dir: str
        :param quality: Quality of the stream to download in the format [resolution]p[framerate]
        :type quality: str
        :param quiet: True suppresses progress reporting
        :type quiet: bool
        """
        super().__init__(parent_dir, quiet)

        self.__setattr__("_quality", quality)

        # create segments for combining with archived VOD parts. If true we will try to recreate the segment numbering
        # scheme Twitch uses, otherwise we use our own numbering scheme. Only used when archiving a live stream without
        # a VOD.
        self._align_segments: bool = True

        # buffers and progress tracking
        self._index_uri: str = ""
        self._incoming_part_buffer: list[StreamSegment.Part] = []
        self._download_queue: StreamSegmentList = None
        self._completed_segments: set[StreamSegment] = set()
        self._processed_parts: set[StreamSegment.Part] = set()
        self._last_part_announce: float = datetime.now(timezone.utc).timestamp()

        self._unsupported_parts = set()

        # channel-specific vars
        self.channel: Channel = channel
        self.output_dir: Path = None
        self.vod: Vod = vod

        # perform setup
        self._do_setup()

    def __repr__(self):
        return str(
            {"channel": self.channel, "index_uri": self._index_uri, "stream": self.vod}
        )

    def export_metadata(self):
        write_json_file(self.vod.to_dict(), Path(self.output_dir, "vod.json"))

    def start(self):
        """
        Begins downloading the stream for the channel until stopped or stream ends.
        """
        # create output  dir if not already created
        Path(self.output_dir, "parts").mkdir(parents=True, exist_ok=True)

        # loop until stream ends
        while True:
            _start_timestamp: float = datetime.now(timezone.utc).timestamp()
            self.single_download_pass()

            # assume stream has ended once >20s has passed since the last segment was advertised
            #   if parts remain in the buffer, we need to download them whether there are 5 parts or not
            if time_since_date(self._last_part_announce) > 20:
                # perform secondary check to see if stream is actually offline
                _stream_info = self.channel.get_stream_info()
                # check channel stream id matches ours
                if (
                    _stream_info["stream"]
                    and int(_stream_info["stream"]["id"]) == self.vod.s_id
                ):
                    pass

                else:
                    self._log.debug(
                        "Assuming stream has ended as 20 seconds passed since last segment announced."
                    )
                    self._get_final_segment()
                    break

            # sleep if processing time < CHECK_INTERVAL time before checking for new segments
            _loop_time = int(datetime.now(timezone.utc).timestamp() - _start_timestamp)
            if _loop_time < CHECK_INTERVAL:
                sleep(CHECK_INTERVAL - _loop_time)

    def merge(self):
        """
        Attempt to merge downloaded segments.
        """
        # we pass all possible segments as muted segments so that corrupt parts are ignored.
        merger = Merger(
            self.vod,
            self.output_dir,
            self._completed_segments,
            [MpegSegment(i, 10) for i in range(10000)],
            self._quiet,
        )

        # attempt to merge - no verification done as unsynced streams have no accurate duration to verify against
        try:
            merger.merge()
            merger.cleanup_temp_files()

        except Exception as exc:
            raise VideoMergeError("Exception raised while merging VOD.") from exc

    def single_download_pass(self):
        """
        Used to fetch and download stream segments without looping. This is used for creating a stream buffer at the
        start of archiving in case a VOD never becomes available, in which case the previously broadcast segments
        would be lost.
        """
        try:
            self._fetch_advertised_parts()

            if self._incoming_part_buffer:
                self._build_download_queue()
                self._download_queued_segments()

        # stream offline
        except TwitchAPIErrorNotFound:
            self._log.info("%s is offline or stream ended.", self.channel.name)
            if self._download_queue:
                self._get_final_segment()

    def _do_setup(self):
        """
        Performs required setup prior to starting download.
        """
        # generate VOD from channel stream information if not provided -
        # channel stream information can differ from video / VOD information making parallel archiving with the video
        # downloader impossible, so we need the option to provide our own VOD to sync with.
        if not self.vod:
            self._log.debug("Fetching required stream information.")
            # retry loop to avoid StreamOfflineError if stream just went live
            for _ in range(4):
                if _ == 3:
                    self._log.info("%s is offline.", self.channel.name)
                    raise StreamOfflineError(self.channel)

                stream_vod = ArchivedVod.convert_from_vod(
                    Vod.from_stream_json(self.channel.get_stream_info()),
                    video_archived=True,
                )
                stream_vod.channel = self.channel
                if not stream_vod:
                    sleep(5)
                    continue

                self.vod = stream_vod
                break

        # fetch index
        try:
            self._index_uri = self.channel.get_stream_index(self._quality)
        except TwitchAPIErrorNotFound as exc:
            raise StreamOfflineError(self.channel) from exc

        # ensure enough time has passed for VOD api to update before archiving. this is important as it changes
        # the way we archive if the stream is not being archived to a VOD.
        self._log.debug("Current stream length: %s", self.vod.duration)

        # while we wait for the api to update we must build a temporary buffer of any parts advertised in the
        # meantime in case there is no vod and thus no way to retrieve them after the fact
        if not self.vod.v_id:
            if self.vod.duration < TEMP_BUFFER_LEN:
                self._buffer_stream(self.vod.duration)

            # fetch most recent channel broadcast ID
            broadcast_vod_id = self.channel.broadcast_v_id

            # if broadcast ID stream_id matches ours, they are paired.
            broadcast_vod = Vod(broadcast_vod_id)
            if broadcast_vod.s_id == self.vod.s_id:
                self.vod = Vod(broadcast_vod_id)

                # if a paired VOD exists for the stream we can discard our temporary buffer and any downloaded files
                self._incoming_part_buffer: list[StreamSegment.Part] = []
                self._processed_parts: set[StreamSegment.Part] = set()
                self._last_part_announce: float = datetime.now(timezone.utc).timestamp()

                if self.output_dir and Path(self.output_dir).exists():
                    shutil.rmtree(Path(self.output_dir))

            # stream is not paired with broadcast VOD - must be stream-only
            else:
                self._align_segments = False

        # build and create actual output directory
        self.output_dir = Path(
            self._parent_dir,
            build_output_dir_name(self.vod.title, self.vod.created_at, self.vod.v_id),
        )

        if self.output_dir.exists():
            # get existing parts to resume counting if archiving halted
            self._completed_segments = {
                MpegSegment(int(Path(p).name.removesuffix(".ts")), 10)
                for p in list(Path(self.output_dir, "parts").glob("*.ts"))
            }

        self._init_download_queue()

    def _buffer_stream(self, stream_length: int):
        """
        Builds a temporary buffer when a stream has just started to ensure the Twitch API has updated when we
        attempt to fetch VOD information.

        :param stream_length: time in seconds stream has been live
        """
        self._log.debug(
            "Stream began less than 120s ago, delaying archival start until VOD API updated."
        )

        self._init_download_queue()

        # create temporary download directory
        self.output_dir = Path(
            self._parent_dir, build_output_dir_name(self.vod.title, self.vod.created_at)
        )
        Path(self.output_dir, "parts").mkdir(parents=True, exist_ok=True)

        # download new parts every CHECK_INTERVAL time
        for _ in range(int((TEMP_BUFFER_LEN - stream_length) / CHECK_INTERVAL)):
            _start_timestamp: float = datetime.now(timezone.utc).timestamp()
            self.single_download_pass()

            # wait if less than CHECK_INTERVAL time passed since grabbing more parts
            _loop_time = int(datetime.now(timezone.utc).timestamp() - _start_timestamp)
            if _loop_time < CHECK_INTERVAL:
                sleep(CHECK_INTERVAL - _loop_time)

        # wait any remaining time
        sleep((TEMP_BUFFER_LEN - stream_length) % CHECK_INTERVAL)

    def _init_download_queue(self):
        # start at '-1', the first part will then be '0'
        _latest_segment = MpegSegment(-1, 0)
        if self._completed_segments:
            # set start segment for download queue
            _latest_segment = max(self._completed_segments, key=attrgetter("id"))

        # pass stream created to segment list to be used for determining part and segment ids
        self._download_queue = StreamSegmentList(
            self.vod.created_at, self._align_segments, _latest_segment.id
        )

    def _fetch_advertised_parts(self):
        """
        Fetch parts being advertised by Twitch for stream.
        """
        # attempt to grab new parts from Twitch
        for _ in range(6):
            try:
                if _ >= 5:
                    raise StreamFetchError(
                        self.channel.name,
                        "Request timed out while fetching new segments.",
                    )

                # fetch advertised stream parts
                announced_parts = m3u8.loads(
                    self.channel.get_stream_playlist(self._index_uri)
                ).segments
                self._last_part_announce = (
                    announced_parts[-1]
                    .program_date_time.replace(tzinfo=timezone.utc)
                    .timestamp()
                )

                for _part in [StreamSegment.Part(_p) for _p in announced_parts]:
                    # add new parts to part buffer
                    if _part not in self._processed_parts:
                        self._processed_parts.add(_part)
                        self._incoming_part_buffer.append(_part)
                        self.vod.duration = int(_part.timestamp - self.vod.created_at)
                break

            # retry if request times out
            except RequestError as exc:
                self._log.debug(
                    "Error attempting to fetch new stream segments. (Attempt %s). Error: %s",
                    _ + 1,
                    exc,
                )
                continue

    def _build_download_queue(self):
        """
        Creates queue of segments being downloaded using the incoming part buffer and already processed segments.
        """
        # add parts to the associated segment
        for _part in self._incoming_part_buffer:
            if _part.title != "live":
                self._log.debug(
                    "Ignoring advertisement part %s.",
                    {
                        "url": "(hash)" + str(hash(_part.url)),
                        "timestamp": _part.timestamp,
                        "duration": _part.duration,
                    },
                )
                continue

            # some streams have part lengths other than the default of 2.0. these cannot be aligned, and so we raise
            # an error if we encounter more than one if we are attempting to align the segments. we check for >2
            # instead of just 1 as the final part (or two) in the stream is often shorter than 2.0.
            if self._align_segments and _part.duration != 2.0:
                self._log.debug(
                    "Found part with unsupported duration (%s).", _part.duration
                )
                self._unsupported_parts.add(_part)

            if len(self._unsupported_parts) > 2:
                raise UnsupportedStreamPartDuration

            # add part to segment download queue
            self._log.debug(
                "Adding part %s to download queue.",
                {
                    "url": "(hash)" + str(hash(_part.url)),
                    "timestamp": _part.timestamp,
                    "duration": _part.duration,
                },
            )
            self._download_queue.add_part(_part)

        # wipe part buffer
        self._incoming_part_buffer = []

    def _download_queued_segments(self):
        """
        Downloads all queued segments.
        """
        for _segment_id in self._download_queue.get_completed_segment_ids():
            self._download_segment(self._download_queue.pop_segment(_segment_id))

    def _download_segment(self, segment: StreamSegment):
        """
        Downloads a given segment.

        :param segment: StreamSegment to download.
        """
        # generate buffer file path
        _temp_buffer_file = Path(
            tempfile.gettempdir(),
            "twitch-archiver",
            str(self.vod.s_id),
            str(f"{segment.id:05d}" + ".ts"),
        )
        _temp_buffer_file.parent.mkdir(parents=True, exist_ok=True)

        # begin retry loop for download
        for _ in range(6):
            if _ >= 5:
                self._log.error(
                    "Maximum attempts reached while downloading segment %s.", segment.id
                )
                return

            self._log.debug(
                "Downloading segment %s to %s.", segment.id, _temp_buffer_file
            )
            with open(_temp_buffer_file, "wb") as _tmp_file:
                # iterate through each part of the segment, downloading them in order
                for _part in segment.parts:
                    try:
                        _r = requests.get(_part.url, stream=True, timeout=5)

                        if _r.status_code != 200:
                            return

                        # write part to file
                        for chunk in _r.iter_content(chunk_size=262144):
                            _tmp_file.write(chunk)

                    except RequestError as exc:
                        self._log.debug(
                            "Error downloading stream segment %s: %s",
                            segment.id,
                            str(exc),
                        )

            # move finished ts file to destination storage
            try:
                safe_move(
                    Path(_temp_buffer_file),
                    Path(self.output_dir, "parts", str(f"{segment.id:05d}" + ".ts")),
                )
                self._completed_segments.add(segment)
                self._log.debug("Stream segment: %s completed.", segment.id)
                break

            except Exception as exc:
                raise StreamSegmentDownloadError(segment.id, self.channel.name) from exc

    def _get_final_segment(self):
        """
        Downloads the final stream segment.
        """
        # ensure final segment present
        if self._download_queue.is_segment_present(self._download_queue.current_id):
            self._log.debug("Fetching final stream segment.")
            self._download_segment(
                self._download_queue.get_segment_by_id(self._download_queue.current_id)
            )

    def cleanup_temp_files(self):
        """
        Deletes all temporary files and directories.
        """
        shutil.rmtree(Path(self.output_dir, "parts"), ignore_errors=True)
        if self.vod.v_id:
            shutil.rmtree(
                Path(tempfile.gettempdir(), "twitch-archiver", str(self.vod.v_id)),
                ignore_errors=True,
            )
        else:
            shutil.rmtree(
                Path(tempfile.gettempdir(), "twitch-archiver", str(self.vod.s_id)),
                ignore_errors=True,
            )



"""
Module used for downloading chat logs for a given Twitch VOD.
"""
import json
import os
from datetime import datetime, timezone
from pathlib import Path
from time import sleep

from twitcharchiver.api import Api
from twitcharchiver.downloader import Downloader
from twitcharchiver.exceptions import (
    TwitchAPIErrorNotFound,
    RequestError,
    ChatDownloadError,
    TwitchAPIErrorForbidden,
)
from twitcharchiver.utils import (
    Progress,
    get_time_difference,
    write_json_file,
    write_file_line_by_line,
    build_output_dir_name,
    time_since_date,
    parse_twitch_timestamp,
)
from twitcharchiver.vod import Vod, ArchivedVod

CHECK_INTERVAL = 60


class Chat(Downloader):
    """
    Class which handles the downloading, updating and importing of chat logs and the subsequent fomatting and archival.
    """

    def __init__(
        self, vod: Vod, parent_dir: Path = Path(os.getcwd()), quiet: bool = False
    ):
        """
        Initialize class variables.

        :param vod: VOD to be downloaded
        :param parent_dir: path to parent directory for downloaded files
        :param quiet: boolean whether to print progress
        """
        # init downloader
        super().__init__(parent_dir, quiet)

        # setup api with required header
        self._api = Api()
        self._api.add_headers({"Client-Id": "ue6666qo983tsx6so1t0vnawi233wa"})

        # vod-specific vars
        self.vod: Vod = vod
        # create output dir
        self.output_dir = Path(
            self._parent_dir,
            build_output_dir_name(self.vod.title, self.vod.created_at, self.vod.v_id),
        )

        # store chat log and seen message ids
        self._chat_log: list = []
        self._chat_message_ids: set = set()

        # load chat from file if a download was attempted previously
        self.load_from_file()

    def export_metadata(self):
        write_json_file(self.vod.to_dict(), Path(self.output_dir, "vod.json"))

    def load_from_file(self):
        """
        Loads the chat log stored in the output directory.
        """
        try:
            with open(
                Path(self.output_dir, "verbose_chat.json"), "r", encoding="utf8"
            ) as chat_file:
                self._log.debug("Loading chat log from file.")
                chat_log = json.loads(chat_file.read())

            # ignore chat logs created with older incompatible schema - see v2.2.1 changes
            if chat_log and "contentOffsetSeconds" not in chat_log[0].keys():
                self._log.debug(
                    "Ignoring chat log loaded from file as it is incompatible."
                )
                return
            self._log.debug("Chat log found for VOD %s.", self.vod)
            self._chat_log = chat_log
            # add chat messages to id set
            self._chat_message_ids.update(m["id"] for m in chat_log)
            return

        except FileNotFoundError:
            return

    def start(self):
        """
        Downloads the chat for the given VOD and exports both a readable and JSON-formatted log to the provided
        directory.

        :return: list of dictionaries containing chat message data
        :rtype: list[dict]
        """
        try:
            # create output dir
            Path(self.output_dir).mkdir(parents=True, exist_ok=True)

            self._download()
            self.export_chat_logs()

            # use while loop for archiving live VODs
            while self.vod.is_live():
                self._log.debug(
                    "VOD is still live, attempting to download new chat segments."
                )
                _start_timestamp: float = datetime.now(timezone.utc).timestamp()

                # refresh VOD metadata
                self.vod.refresh_vod_metadata()

                # begin downloader from offset of previous log
                if self._chat_log:
                    self._download(self._chat_log[-1]["contentOffsetSeconds"])
                else:
                    self._download()
                self.export_chat_logs()

                # sleep if processing time < CHECK_INTERVAL before fetching new messages
                _loop_time = int(
                    datetime.now(timezone.utc).timestamp() - _start_timestamp
                )
                if _loop_time < CHECK_INTERVAL:
                    sleep(CHECK_INTERVAL - _loop_time)

            # delay final archive pass if stream just ended
            self.vod.refresh_vod_metadata()
            if time_since_date(self.vod.created_at + self.vod.duration) < 300:
                self._log.debug(
                    "Stream ended less than 5m ago, delaying before final chat archive attempt."
                )
                sleep(300)

                # refresh VOD metadata
                self.vod.refresh_vod_metadata()
                if self._chat_log:
                    self._download(self._chat_log[-1]["contentOffsetSeconds"])
                else:
                    self._download()

        except (TwitchAPIErrorNotFound, TwitchAPIErrorForbidden):
            self._log.debug(
                "Error 403 or 404 returned when checking for new chat segments - VOD was likely deleted."
            )

        finally:
            self.export_chat_logs()

        # logging
        self._log.info("Found %s chat messages.", len(self._chat_log))

        # set archival flag if ArchivedVod provided
        if isinstance(self.vod, ArchivedVod):
            self.vod.chat_archived = True

    def _download(self, offset: int = 0):
        """
        Downloads the chat log in its entirety.

        :param offset: time to begin archiving from in seconds
        :return: list of all chat messages
        :rtype: list
        """
        _progress = Progress()
        start_len = len(self._chat_log)

        # grab initial chat segment containing cursor
        _initial_segment, _cursor = self._get_chat_segment(offset=offset)
        self._chat_log.extend(
            [m for m in _initial_segment if m["id"] not in self._chat_message_ids]
        )
        self._chat_message_ids.update([m["id"] for m in _initial_segment])

        while True:
            if not _cursor:
                self._log.debug(
                    f"{len(self._chat_log) - start_len} messages retrieved from Twitch."
                )
                break

            self._log.debug("Fetching chat segments at cursor: %s.", _cursor)
            # grab next chat segment along with cursor for next segment
            _segment, _cursor = self._get_chat_segment(cursor=_cursor)
            self._chat_log.extend(
                [m for m in _segment if m["id"] not in self._chat_message_ids]
            )
            self._chat_message_ids.update([m["id"] for m in _segment])
            # vod duration in seconds is used as the total for progress bar
            # comment offset is used to track what's been done
            # could be done properly if there was a way to get the total number of comments
            if not self._quiet:
                _progress.print_progress(
                    int(_segment[-1]["contentOffsetSeconds"]), self.vod.duration
                )

    def _get_chat_segment(self, offset: int = 0, cursor: str = ""):
        """
        Retrieves a chat segment and any subsequent segments from a given offset.

        :param offset: offset in seconds to begin retrieval from
        :type offset: int
        :param cursor: cursor returned by a previous call of this function
        :type cursor: str
        :returns: list of comments, cursor if one is returned from twitch
        :rtype: list, str
        """
        # build payload
        if offset != 0:
            _p = [
                {
                    "operationName": "VideoCommentsByOffsetOrCursor",
                    "variables": {
                        "videoID": str(self.vod.v_id),
                        "contentOffsetSeconds": offset,
                    },
                }
            ]

        else:
            _p = [
                {
                    "operationName": "VideoCommentsByOffsetOrCursor",
                    "variables": {"videoID": str(self.vod.v_id), "cursor": cursor},
                }
            ]

        _p[0]["extensions"] = {
            "persistedQuery": {
                "version": 1,
                "sha256Hash": "b70a3591ff0f4e0313d126c6a1502d79a1c02baebb288227c582044aa76adf6a",
            }
        }

        for _ in range(5):
            try:
                _r = self._api.post_request("https://gql.twitch.tv/gql", j=_p).json()
                _comments = _r[0]["data"]["video"]["comments"]

                # check if next page exists
                if _comments:
                    if _comments["pageInfo"]["hasNextPage"]:
                        return [c["node"] for c in _comments["edges"]], _comments[
                            "edges"
                        ][-1]["cursor"]

                    return [c["node"] for c in _comments["edges"]], None

                return [], None

            except RequestError:
                self._log.error(
                    f"Error downloading chat segment at cursor or offset: {cursor or offset}"
                )
                continue

        raise ChatDownloadError(
            f"Maximum attempts reached while downloading chat segment at cursor or offset: {cursor or offset}"
        )

    def generate_readable_chat_log(self, chat_log: list):
        """
        Converts the raw chat log into a human-readable format.

        :param chat_log: list of messages to generate log from
        :type chat_log: list
        """
        _r_chat_log = []
        for _comment in chat_log:
            _created_time = parse_twitch_timestamp(_comment["createdAt"])

            _comment_time = (
                f"{get_time_difference(self.vod.created_at, _created_time):.3f}"
            )

            # catch comments without commenter information
            if _comment["commenter"]:
                _user_name = str(_comment["commenter"]["displayName"])
            else:
                _user_name = "~MISSING_COMMENTER_INFO~"

            # catch comments without data
            if _comment["message"]["fragments"]:
                _user_message = str(_comment["message"]["fragments"][0]["text"])
            else:
                _user_message = "~MISSING_MESSAGE_INFO~"

            _user_badges = ""
            try:
                for _badge in _comment["message"]["userBadges"]:
                    if "broadcaster" in _badge["setID"]:
                        _user_badges += "(B)"

                    if "moderator" in _badge["setID"]:
                        _user_badges += "(M)"

                    if "subscriber" in _badge["setID"]:
                        _user_badges += "(S)"

            except KeyError:
                pass

            # FORMAT: [TIME] (B1)(B2)NAME: MESSAGE
            _r_chat_log.append(
                f"[{_comment_time}] {_user_badges}{_user_name}: {_user_message}"
            )

        return _r_chat_log

    def export_chat_logs(self):
        """
        Exports a readable and a JSON-formatted chat log to the output directory.
        """
        write_file_line_by_line(
            self.generate_readable_chat_log(self._chat_log),
            Path(self.output_dir, "readable_chat.txt"),
        )
        write_json_file(self._chat_log, Path(self.output_dir, "verbose_chat.json"))

    def get_message_count(self):
        """
        Fetches the total number of retrieved chat messages.

        :return:
        """
        return len(self._chat_log)
